#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import re
import os
import uuid
import logging
import tempfile
import requests

import torch

from pathlib import Path
from typing import Iterable, List, Optional, Tuple
from urllib.parse import urlparse
from azureml.core import Run

from health_azure import download_checkpoints_from_run_id, get_workspace, torch_barrier
from health_azure.utils import (
    RUN_CONTEXT,
    _download_files_from_run,
    get_run_file_names,
    is_running_in_azure_ml,
    is_local_rank_zero,
    download_files_by_suffix,
)
from health_ml.utils.common_utils import (
    AUTOSAVE_CHECKPOINT_CANDIDATES,
    CHECKPOINT_FOLDER,
    DEFAULT_AML_CHECKPOINT_DIR,
    CHECKPOINT_SUFFIX,
    DEFAULT_AML_UPLOAD_DIR,
)
from health_ml.utils.type_annotations import PathOrString

# This is a constant that must match a filename defined in pytorch_lightning.ModelCheckpoint, but we don't want
# to import that here.
LAST_CHECKPOINT_FILE_NAME = f"last{CHECKPOINT_SUFFIX}"
LEGACY_RECOVERY_CHECKPOINT_FILE_NAME = "recovery"
MODEL_INFERENCE_JSON_FILE_NAME = "model_inference_config.json"
MODEL_WEIGHTS_DIR_NAME = "pretrained_models"

# The dictionary field where PyTorch Lightning stores the epoch number in the checkpoint file.
CHECKPOINT_EPOCH_KEY = "epoch"

# The string that is used to prefix the folders with results from retries after low priority preemption in AzureML.
AZUREML_RETRY_PREFIX = "retry_"


def get_best_checkpoint_path(path: Path) -> Path:
    """
    Given a path and checkpoint, formats a path based on the checkpoint file name format.

    :param path to checkpoint folder
    """
    return path / LAST_CHECKPOINT_FILE_NAME


def download_folder_from_run_to_temp_folder(
    folder: str, run: Optional[Run] = None, run_barrier: bool = True  # type: ignore
) -> Path:
    """
    Downloads all files from a run that have the given prefix to a temporary folder.
    For example, if the run contains files "foo/bar.txt" and "nothing.txt", and this function is called with
    argument folder = "foo", it will return a path in a temp file system pointing to "bar.txt".

    In distributed training, the download only happens once per node, on local rank zero. All ranks wait for the
    completion of the downloads, by calling `torch.barrier()`. This is to avoid multiple ranks trying to download
    the same files at the same time. Calling the barrier can be skipped by setting `run_barrier` to False.

    :param folder: The folder to download. This is the prefix of the files to download.
    :param run: If provided, download the files from that run. If omitted, download the files from the current run
        (taken from RUN_CONTEXT)
    :param run_barrier: If True, call torch.barrier() after the download.
    :return: The path to which the files were downloaded. The files are located in that folder, without any further
        subfolders.
    """
    if not is_running_in_azure_ml() and run is None:
        raise ValueError("When running outside AzureML, the run to download from must be set.")
    run: Run = run or RUN_CONTEXT  # type: ignore
    temp_folder = Path(tempfile.mkdtemp())
    cleaned_prefix = folder.strip("/") + "/"
    existing_checkpoints = get_run_file_names(run, prefix=cleaned_prefix)
    logging.info(f"Number of checkpoints available in AzureML: {len(existing_checkpoints)}")
    if len(existing_checkpoints) > 0:
        try:
            logging.info(f"Downloading checkpoints to {temp_folder}")
            _download_files_from_run(run=run, output_dir=temp_folder, prefix=cleaned_prefix)  # type: ignore
        except Exception as ex:
            logging.warning(f"Unable to download checkpoints from AzureML. Error: {str(ex)}")
    # Checkpoint downloads preserve the full folder structure, point the caller directly to the folder where the
    # checkpoints really are.
    if run_barrier:
        torch_barrier()
    return temp_folder / cleaned_prefix


def _get_checkpoint_files(files: List[str]) -> List[str]:
    """Filters a list of files in an AzureML run to only retain those that could be recovery or last checkpoints.
    This takes the folder structures for retries into account, where files are written into subfolders like
    `retry_001`.

    :param files: The list of file names to check.
    :return: A list of file names that could be recovery checkpoints."""
    folder_pattern = f"{DEFAULT_AML_UPLOAD_DIR}(/{AZUREML_RETRY_PREFIX}[0-9]" + "{3})?" + f"/{CHECKPOINT_FOLDER}/"
    checkpoint_filenames = (*AUTOSAVE_CHECKPOINT_CANDIDATES, LAST_CHECKPOINT_FILE_NAME)
    file_pattern = "|".join(f"({filename})" for filename in checkpoint_filenames)
    pattern = f"{folder_pattern}({file_pattern})"
    regex = re.compile(pattern)
    return [f for f in files if regex.match(f)]


def download_checkpoints_from_run(run: Run, tmp_folder: Optional[Path] = None) -> Path:
    """This function will download all checkpoints from the current AzureML run to a temporary folder. It will take
    into account cases where a job got pre-empted, and wrote checkpoints to subfolders like `retry_001`.

    In distributed training, the download only happens once per node, on local rank zero. All ranks wait for the
    completion of the downloads, by calling `torch.barrier()`. This is to avoid multiple ranks trying to download
    the same files at the same time.

    :param run: The AzureML run to download the checkpoints from.
    :param tmp_folder: The folder to download the checkpoints to. If None, a temporary folder will be created.
    :return: The folder to which the checkpoints were downloaded."""
    tmp_folder = tmp_folder or Path(tempfile.mkdtemp())
    if is_local_rank_zero():
        for file in _get_checkpoint_files(run.get_file_names()):
            logging.info(f"Downloading checkpoint file {file} to temp folder")
            run.download_file(file, output_file_path=str(tmp_folder / file))
    torch_barrier()
    return tmp_folder


def download_highest_epoch_checkpoint(run: Run, checkpoint_suffix: str, output_folder: Path) -> Optional[Path]:
    """Downloads all checkpoint files from the run that have a given suffix, and returns the local path to the
    downloaded checkpoint with the highest epoch. Checkpoints are downloaded one-by-one and deleted right away
    if they are not the highest epoch.

    :param run: The AzureML run from where the files should be downloaded.
    :param checkpoint_suffix: The suffix for all files that should be returned.
    :param output_folder: The folder to download the checkpoints to.
    :return: The path to the downloaded file that has the highest epoch. Returns None if no suitable checkpoint
        was found, or if the epoch information could not be extracted.
    """
    files_from_run = download_files_by_suffix(
        run=run, suffix=str(checkpoint_suffix), output_folder=output_folder, validate_checksum=True
    )
    return find_checkpoint_with_highest_epoch(files_from_run, delete_files=True)


def find_recovery_checkpoint_on_disk_or_cloud(path: Path) -> Optional[Path]:
    """
    Looks at all the checkpoint files and returns the path to the one that should be used for recovery.
    If no checkpoint files are found on disk, the function attempts to download from the current AzureML
    run.

    :param path: The folder to start searching in.
    :return: None if there is no suitable recovery checkpoints, or else a full path to the checkpoint file.
    """
    recovery_checkpoint = find_local_recovery_checkpoint(path)
    if recovery_checkpoint is None and is_running_in_azure_ml():
        logging.info(
            "No recovery checkpoints available in the checkpoint folder. Trying to find checkpoints in AzureML."
        )
        # Download checkpoints from AzureML, then try to find recovery checkpoints among those.
        # Downloads should go to a temporary folder because downloading the files to the checkpoint
        # folder might cause artifact conflicts later.
        temp_folder = download_checkpoints_from_run(run=RUN_CONTEXT)
        recovery_checkpoint = find_recovery_checkpoint_in_downloaded_files(temp_folder)
    return recovery_checkpoint


def _load_epoch_from_checkpoint(path: Optional[Path]) -> int:
    """
    Loads the epoch number from the given checkpoint file.

    :param path: Path to checkpoint file
    """
    if path is None:
        raise ValueError("Checkpoint path is None")
    checkpoint = torch.load(str(path), map_location=torch.device("cpu"))
    return checkpoint[CHECKPOINT_EPOCH_KEY]


def find_checkpoint_with_highest_epoch(files: Iterable[Path], delete_files: bool = False) -> Optional[Path]:
    """Loads the epoch numbers from the given checkpoint files, and returns the one with the
    highest epoch number. If no files can be loaded, or the list is empty, None is returned.
    If the `delete_files` flag is set to True, all files apart from the one with the highest epoch are deleted.

    :param files: A list of checkpoint files.
    :param delete_files: If True, all files apart from the one with the highest epoch are deleted. If False,
        no files are deleted.
    :return: The checkpoint file with the highest epoch number, or None if no such file was found."""

    def update_file_with_highest_epoch(
        file: Path, highest_epoch: Optional[int], file_with_highest_epoch: Optional[Path]
    ) -> Tuple[int, Path]:
        """Reads the epoch number from the given `file`, and returns updated information about which file
        has been found to have the highest epoch number.

        :param file: The name of the file to process.
        :param highest_epoch: The highest epoch number encountered so far, or None if no files processed yet.
        :param file_with_highest_epoch: The file that contained the highest epoch number so far, or None if no
            files processed yet.
        :return: A tuple of updated (`highest_epoch`, `file_with_highest_epoch`) values.
        """
        epoch = _load_epoch_from_checkpoint(file)
        logging.info(f"Checkpoint for epoch {epoch} in {file}")
        if (highest_epoch is None) or (epoch > highest_epoch):
            if file_with_highest_epoch is not None and delete_files:
                logging.debug(f"Deleting checkpoint file {file_with_highest_epoch}")
                file_with_highest_epoch.unlink()
            return epoch, file
        assert file_with_highest_epoch is not None
        return highest_epoch, file_with_highest_epoch

    highest_epoch: Optional[int] = None
    file_with_highest_epoch: Optional[Path] = None
    for file in files:
        if file.is_file():
            try:
                highest_epoch, file_with_highest_epoch = update_file_with_highest_epoch(
                    file, highest_epoch, file_with_highest_epoch
                )
            except Exception as ex:
                logging.warning(f"Unable to handle checkpoint file {file}: {ex}")
    return file_with_highest_epoch


def find_local_recovery_checkpoint(path: Path) -> Optional[Path]:
    """Checks for a recovery checkpoint in the local checkpoint folder. This can be either
    an autosave checkpoint or the last checkpoint.

    :param path: The folder to search in.
    :return: The checkpoint file with the highest epoch number, or None if no such file was found.
    """
    candidates = [path / f for f in (*AUTOSAVE_CHECKPOINT_CANDIDATES, LAST_CHECKPOINT_FILE_NAME)]
    return find_checkpoint_with_highest_epoch(candidates)


def find_recovery_checkpoint_in_downloaded_files(path: Path) -> Optional[Path]:
    """
    Finds the checkpoint file in the given path that can be used for re-starting the present job.
    All existing checkpoints are loaded, and the one for the highest epoch is used for recovery.

    :param path: The folder to search in.
    :return: The checkpoint file with the highest epoch number, or None if no such file was found.
    """
    candidates = path.glob(f"**/*{CHECKPOINT_SUFFIX}")
    return find_checkpoint_with_highest_epoch(candidates, delete_files=False)


def cleanup_checkpoints(ckpt_folder: Path) -> None:
    """
    Remove autosave checkpoints from the given checkpoint folder, and check if a "last.ckpt" checkpoint is present.

    :param ckpt_folder: The folder that contains all checkpoint files.
    """
    files_in_checkpoint_folder = [p.name for p in ckpt_folder.glob('*')]
    if len(files_in_checkpoint_folder) == 0:
        return
    logging.info(f"Files in checkpoint folder: {' '.join(files_in_checkpoint_folder)}")
    last_ckpt = ckpt_folder / LAST_CHECKPOINT_FILE_NAME
    all_files = f"Existing files: {' '.join(p.name for p in ckpt_folder.glob('*'))}"
    if not last_ckpt.is_file():
        raise FileNotFoundError(f"Checkpoint file {LAST_CHECKPOINT_FILE_NAME} not found. {all_files}")
    # Training is finished now. To save storage, remove the autosave checkpoint which is now obsolete.
    # Lightning does not overwrite checkpoints in-place. Rather, it writes "autosave.ckpt",
    # then "autosave-1.ckpt" and deletes "autosave.ckpt", then "autosave.ckpt" and deletes "autosave-v1.ckpt"
    for candidate in AUTOSAVE_CHECKPOINT_CANDIDATES:
        autosave = ckpt_folder / candidate
        if autosave.is_file():
            autosave.unlink()


class CheckpointDownloader:
    def __init__(
        self,
        run_id: str,
        checkpoint_filename: Optional[str] = None,
        download_dir: PathOrString = "checkpoints",
        remote_checkpoint_dir: Optional[Path] = None,
    ) -> None:
        """
        Utility class for downloading checkpoint files from an Azure ML run

        :param run_id: Recovery ID of the run from which to load the checkpoint.
        :param checkpoint_filename: Name of the checkpoint file, expected to be inside the
        `outputs/checkpoints/` directory (e.g. `"best_checkpoint.ckpt"`).
        :param download_dir: The local directory in which to save the downloaded checkpoint files.
        :param remote_checkpoint_dir: The remote folder from which to download the checkpoint file
        """
        self.run_id = run_id
        self.checkpoint_filename = checkpoint_filename or self.extract_checkpoint_filename_from_run_id()
        self.download_dir = Path(download_dir)
        self.remote_checkpoint_dir = (
            remote_checkpoint_dir or self.extract_remote_checkpoint_dir_from_checkpoint_filename()
        )

    def extract_checkpoint_filename_from_run_id(self) -> str:
        """
        Extracts the checkpoint filename from the run_id if run_id is in the format
        <MyContainer_xx>:<checkpoint_filename.ckpt>. Otherwise, uses the last epoch checkpoint filename as default.
        """
        run_id_split = self.run_id.split(":")
        self.run_id = run_id_split[0]
        return run_id_split[-1] if len(run_id_split) > 1 else LAST_CHECKPOINT_FILE_NAME

    def extract_remote_checkpoint_dir_from_checkpoint_filename(self) -> Path:
        """
        Extracts the remote checkpoint directory from the checkpoint filename if checkpoint_filename is in the format
        <custom/patch/checkpoint_filename.ckpt>. Otherwise, uses the default remote checkpoint directory
        'outputs/checkpoints/'.
        """
        tmp_checkpoint_filename = self.checkpoint_filename
        checkpoint_filename_split = self.checkpoint_filename.split("/")
        self.checkpoint_filename = checkpoint_filename_split[-1]
        return (
            Path(tmp_checkpoint_filename).parent
            if len(checkpoint_filename_split) > 1
            else Path(DEFAULT_AML_CHECKPOINT_DIR)
        )

    @property
    def local_checkpoint_dir(self) -> Path:
        return self.download_dir / self.run_id

    @property
    def remote_checkpoint_path(self) -> Path:
        assert self.checkpoint_filename is not None
        return self.remote_checkpoint_dir / self.checkpoint_filename

    @property
    def local_checkpoint_path(self) -> Path:
        return self.local_checkpoint_dir / self.remote_checkpoint_path

    def download_checkpoint_if_necessary(self) -> None:
        """Downloads the specified checkpoint if it does not already exist."""

        if not self.local_checkpoint_path.exists():
            workspace = get_workspace()
            self.local_checkpoint_dir.mkdir(exist_ok=True, parents=True)
            download_checkpoints_from_run_id(
                self.run_id, str(self.remote_checkpoint_path), self.local_checkpoint_dir, aml_workspace=workspace
            )
            assert self.local_checkpoint_path.exists(), f"Couln't download checkpoint from run {self.run_id}."


class CheckpointParser:
    """Wrapper class for parsing checkpoint arguments. A checkpoint can be specified in one of the following ways:
    1. A local checkpoint file path
    2. A remote checkpoint file path
    3. A run ID from which to download the checkpoint file
    """

    AML_RUN_ID_FORMAT = (
        f"<AzureML_run_id>:<optional/custom/path/to/checkpoints/><filename{CHECKPOINT_SUFFIX}>"
        f"If no custom path is provided (e.g., <AzureML_run_id>:<filename{CHECKPOINT_SUFFIX}>)"
        "the checkpoint will be downloaded from the default checkpoint folder "
        f"(e.g., '{DEFAULT_AML_CHECKPOINT_DIR}') If no filename is provided, "
        "(e.g., `src_checkpoint=<AzureML_run_id>`) the latest checkpoint "
        f"({LAST_CHECKPOINT_FILE_NAME}) will be downloaded."
    )
    INFO_MESSAGE = (
        "Please provide a valid checkpoint path, URL or AzureML run ID. For custom checkpoint paths "
        f"within an azureml run, provide a checkpoint in the format {AML_RUN_ID_FORMAT}."
    )
    DOC = (
        "We currently support three types of checkpoints: "
        "    a. A local checkpoint folder that contains a checkpoint file."
        "    b. A URL to a remote checkpoint to be downloaded."
        "    c. A previous azureml run id where the checkpoint is supposed to be "
        "       saved ('outputs/checkpoints/' folder by default.)"
        f"For the latter case 'c' : src_checkpoint should be in the format of {AML_RUN_ID_FORMAT}"
    )

    def __init__(self, checkpoint: str = "") -> None:
        self.checkpoint = checkpoint
        self.validate()

    @property
    def is_url(self) -> bool:
        try:
            result = urlparse(self.checkpoint)
            return all([result.scheme, result.netloc])
        except ValueError:
            return False

    @property
    def is_local_file(self) -> bool:
        return Path(self.checkpoint).is_file()

    @property
    def is_aml_run_id(self) -> bool:
        match = re.match(r"[_\w-]*$", self.checkpoint.split(":")[0])
        return match is not None and not self.is_url and not self.is_local_file

    @property
    def is_valid(self) -> bool:
        if self.checkpoint:
            return self.is_local_file or self.is_url or self.is_aml_run_id
        return True

    def validate(self) -> None:
        if not self.is_valid:
            raise ValueError(f"Invalid checkpoint '{self.checkpoint}'. {self.INFO_MESSAGE}")

    @staticmethod
    def download_from_url(url: str, download_folder: Path) -> Path:
        """
        Download a checkpoint from checkpoint_url to the download folder. The file name is determined from
        from the file name in the URL. If that can't be determined, use a random file name.

        :param url: The URL from which to download.
        :param download_folder: The target folder for the download.
        :return: A path to the downloaded file.
        """
        # assign the same filename as in the download url if possible, so that we can check for duplicates
        # If that fails, map to a random uuid
        file_name = os.path.basename(urlparse(url).path) or str(uuid.uuid4().hex)
        checkpoint_path = download_folder / file_name
        # only download if hasn't already been downloaded
        if checkpoint_path.is_file():
            logging.info(f"File already exists, skipping download: {checkpoint_path}")
        else:
            logging.info(f"Downloading from URL {url}")

            response = requests.get(url, stream=True)
            response.raise_for_status()
            with open(checkpoint_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024):
                    file.write(chunk)
        return checkpoint_path

    def get_or_download_checkpoint(self, download_dir: Path) -> Path:
        """Returns the path to the checkpoint file. If the checkpoint is a URL, it will be downloaded to the checkpoints
        folder. If the checkpoint is an AzureML run ID, it will be downloaded from the run to the checkpoints folder.
        If the checkpoint is a local file, it will be returned as is.

        :param download_dir: The checkpoints folder to which the checkpoint should be downloaded if it is a URL or
            AzureML run ID.
        :raises ValueError: If the checkpoint is not a local file, URL or AzureML run ID.
        :raises FileNotFoundError: If the checkpoint is a URL or AzureML run ID and the download fails.
        :return: The path to the checkpoint file.
        """
        if self.is_local_file:
            checkpoint_path = Path(self.checkpoint)
        elif self.is_url:
            download_folder = download_dir / MODEL_WEIGHTS_DIR_NAME
            download_folder.mkdir(exist_ok=True, parents=True)
            checkpoint_path = self.download_from_url(url=self.checkpoint, download_folder=download_folder)
        elif self.is_aml_run_id:
            downloader = CheckpointDownloader(run_id=self.checkpoint, download_dir=download_dir)
            downloader.download_checkpoint_if_necessary()
            checkpoint_path = downloader.local_checkpoint_path
        else:
            raise ValueError("Unable to determine how to get the checkpoint path.")

        if checkpoint_path is None or not checkpoint_path.is_file():
            raise FileNotFoundError(f"Could not find the file at {checkpoint_path}")
        return checkpoint_path



from typing import Any, Callable, Dict, Iterator, List, Mapping, Optional, Sequence

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset, Sampler
from torch.utils.data._utils.collate import default_collate
from math import ceil

from health_ml.utils.common_utils import _create_generator


class BagSampler(Sampler[List[int]]):
    """A batch sampler that iterates samples in specified groups (bags).

    This is useful in multiple-instance learning contexts, where samples in a 'bag' must be
    processed simultaneously. The sampler should be passed to a `DataLoader` as the `batch_sampler`
    argument, for example:

    >>> sampler = BagSampler([1, 1, 0, 2, 0, 1])
    >>> list(sampler)
    [[2, 4], [0, 1, 5], [3]]
    >>> loader = DataLoader(dataset, batch_sampler=sampler)
    """

    def __init__(
        self,
        bag_ids: Sequence,
        shuffle_bags: bool = False,
        shuffle_samples: bool = False,
        max_bag_size: int = 0,
        generator: Optional[torch.Generator] = None,
    ) -> None:
        """
        :param bag_ids: The bag IDs for each sample, of the same length as the dataset.
        :param shuffle_bags: Whether the bags should be iterated in random order.
        :param shuffle_samples: Whether the instances in each bag should be shuffled.
        :param max_bag_size: Upper bound on number of instances in each loaded bag. If 0 (default),
            will return all samples in each bag. If > 0 and `shuffle_samples=True`, bags larger than
            `max_bag_size` will yield random subsets of instances. Note that setting `max_bag_size > 0`
            with `shuffle_samples=False` will return fixed subsets and may completely exclude some
            samples from the iteration.
        :param generator: The pseudorandom number generator to use for shuffling. By default, creates one with a random
            seed.
        """
        self.unique_bag_ids, self.bag_indices = np.unique(bag_ids, return_inverse=True)
        self.shuffle_bags = shuffle_bags
        self.shuffle_samples = shuffle_samples
        self.max_bag_size = max_bag_size
        self.generator = generator

    def __iter__(self) -> Iterator[List[int]]:
        generator = self.generator or _create_generator()
        n_bags = len(self.unique_bag_ids)
        bag_sequence = torch.randperm(n_bags, generator=generator).tolist() if self.shuffle_bags else range(n_bags)
        for bag_idx in bag_sequence:
            yield self.get_bag(bag_idx, generator)

    def get_bag(self, bag_index: int, generator: Optional[torch.Generator] = None) -> List[int]:
        (bag,) = np.where(self.bag_indices == bag_index)
        if self.shuffle_samples:
            if generator is None:
                generator = self.generator or _create_generator()
            perm = torch.randperm(len(bag), generator=generator)
            bag = np.atleast_1d(bag[perm])  # pytorch squeezes singleton tensors
        if self.max_bag_size > 0:
            bag = bag[: self.max_bag_size]
        return bag.tolist()

    def __len__(self) -> int:
        return len(self.unique_bag_ids)

    def __getstate__(self) -> Dict:
        # torch.Generator is not pickleable, so we need to serialise only its internal state.
        # This is useful e.g. for caching and restoring a BagDataset.
        d = self.__dict__.copy()
        if self.generator is not None:
            generator_state = self.generator.get_state() if self.generator is not None else None
            d['generator'] = generator_state
        return d

    def __setstate__(self, d: Dict) -> None:
        # Same here for restoring the torch.Generator state
        self.__dict__ = d
        generator = None
        if d['generator'] is not None:
            generator = torch.Generator()
            generator.set_state(d['generator'])
        self.generator = generator


class BagDataset(Dataset):
    """A wrapper dataset that iterates a base dataset in user-specified 'bags'."""

    def __init__(
        self,
        base_dataset: Sequence,
        bag_ids: Sequence[int],
        shuffle_samples: bool = False,
        max_bag_size: int = 0,
        generator: Optional[torch.Generator] = None,
        collate_fn: Callable[[List], Any] = default_collate,
    ) -> None:
        """
        :param base_dataset: The source dataset whose samples will be grouped in bags.
        :param bag_ids: The bag IDs for each sample, of the same length as the dataset.
        :param shuffle_samples: Whether the instances in each bag should be shuffled.
        :param max_bag_size: Upper bound on number of instances in each loaded bag. If 0 (default),
        will return all samples in each bag. If > 0 and `shuffle_samples=True`, bags larger than
        `max_bag_size` will yield random subsets of instances. Note that setting `max_bag_size > 0`
        with `shuffle_samples=False` will return fixed subsets and may completely exclude some
        samples from the iteration.
        :param generator: The pseudorandom number generator to use for shuffling. By default, creates
        one with a random seed.
        :param collate_fn: Function to aggregate individual samples into a batch. Uses the PyTorch
        default if unspecified, which stacks tensors along their first dimension.
        More details in https://pytorch.org/docs/stable/data.html#dataloader-collate-fn
        """
        if len(base_dataset) != len(bag_ids):
            raise ValueError(
                f"Base dataset and bag IDs must have the same length, " f"got {len(base_dataset)} and {len(bag_ids)}"
            )

        self.base_dataset = base_dataset
        self.bag_sampler = BagSampler(
            bag_ids,
            shuffle_bags=False,  # bag shuffling is handled by dataloader
            shuffle_samples=shuffle_samples,
            max_bag_size=max_bag_size,
            generator=generator,
        )
        self.collate_fn = collate_fn
        self.bag_ids = bag_ids

    def __len__(self) -> int:
        return len(self.bag_sampler)

    def __getitem__(self, index: int) -> Any:
        bag_indices = self.bag_sampler.get_bag(index)
        bag_samples = [self.base_dataset[i] for i in bag_indices]
        return self.collate_fn(bag_samples)


class BatchedDataset(Dataset):
    """A wrapper class that aggregates multiple bags in a batch"""

    # TODO: Just a stub for now; extend to enable shuffling and/or other batching strategies

    def __init__(self, base_dataset: Sequence, batch_size: int) -> None:
        self.base_dataset = base_dataset
        self.batch_size = batch_size

    def __len__(self) -> int:
        return ceil(len(self.base_dataset) / self.batch_size)

    def __getitem__(self, index: int) -> Any:
        start_index = index * self.batch_size
        stop_index = start_index + self.batch_size
        # list of len batch, each element is a dict, dict values are lists or tensors,
        # len of lists/tensors is variable -> we can't use default collate
        multi_bag_list = [self.base_dataset[i] for i in range(start_index, stop_index)]
        return multi_bag_list


def multibag_collate(batch: List) -> Any:
    """Turn list of batches into batch of lists"""
    # TODO: Enable padding tensors to the same fixed dimensions, e.g. using a callable class
    elem = batch[0]

    if isinstance(elem, Mapping):
        elem_keys = elem.keys()
        if not all(other.keys() == elem_keys for other in batch[1:]):
            raise RuntimeError("Every element in the batch should have the same keys")
        return {key: [d[key] for d in batch] for key in elem}

    elif isinstance(elem, Sequence):
        elem_len = len(elem)
        if not all(len(other) == elem_len for other in batch[1:]):
            raise RuntimeError("Every element in the batch should be of equal size")
        return tuple(zip(*batch))

    return batch  # return other types as a plain list


def create_bag_dataloader(
    base_dataset: Sequence,
    bag_ids: Sequence,
    *,  # make following arguments keyword-only to avoid confusion
    shuffle_bags: bool,
    shuffle_samples: bool,
    max_bag_size: int = 0,
    batch_size: Optional[int] = 1,
    collate_fn: Callable[[List], Any] = default_collate,
    generator: Optional[torch.Generator] = None,
    **dataloader_kwargs: Any,
) -> DataLoader:
    """Create a DataLoader that consumes the given dataset in batches grouped by bag ID.

    :param base_dataset: The source dataset whose samples will be grouped in bags.
    :param bag_ids: The bag IDs for each sample, of the same length as the dataset.
    :param shuffle_bags: Whether the bags should be iterated in random order.
    :param shuffle_samples: Whether the instances in each bag should be shuffled.
    :param max_bag_size: Upper bound on number of instances in each loaded bag. If 0 (default),
    will return all samples in each bag. If > 0 and `shuffle_samples=True`, bags larger than
    `max_bag_size` will yield random subsets of instances. Note that setting `max_bag_size > 0`
    with `shuffle_samples=False` will return fixed subsets and may completely exclude some
    samples from the iteration.
    :param batch_size: Number of bags per batch to load. If a number, each field in a batch
    (e.g. dictionary entries or tuple elements) will contain a list of length `batch_size`
    with the corresponding values for every bag. If `None`, will iterate individual bags
    without intermediate lists.
    :param collate_fn: Function to aggregate individual samples into a bag. Uses the PyTorch
    default if unspecified, which stacks tensors along their first dimension.
    More details in https://pytorch.org/docs/stable/data.html#dataloader-collate-fn
    :param generator: The pseudorandom number generator to use for shuffling. By default,
    creates one with a random seed.
    :param dataloader_kwargs: Further keyword arguments to be passed to the `DataLoader`, e.g.
    `num_workers`, `pin_memory`, etc.
    :return: The `DataLoader` configured to iterate one bag at a time.
    """
    generator = generator or _create_generator()
    bag_dataset = BagDataset(
        base_dataset,
        bag_ids=bag_ids,
        max_bag_size=max_bag_size,
        shuffle_samples=shuffle_samples,
        generator=generator,
        collate_fn=collate_fn,
    )
    return DataLoader(
        bag_dataset,
        shuffle=shuffle_bags,
        generator=generator,
        batch_size=batch_size,
        collate_fn=None if batch_size is None else multibag_collate,
        **dataloader_kwargs,
    )



from enum import Enum
from pathlib import Path
from typing import Optional
import param


class DebugDDPOptions(Enum):
    OFF = "OFF"
    INFO = "INFO"
    DETAIL = "DETAIL"


class RunnerMode(Enum):
    TRAIN = "train"
    EVAL_FULL = "eval_full"


class LogLevel(Enum):
    ERROR = "ERROR"
    WARNING = "WARNING"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"


DEBUG_DDP_ENV_VAR = "TORCH_DISTRIBUTED_DEBUG"


class ExperimentConfig(param.Parameterized):
    cluster: str = param.String(
        default="",
        allow_None=False,
        doc="The name of the GPU or CPU cluster inside the AzureML workspace"
        "that should execute the job. To run on your local machine, omit this argument.",
    )
    num_nodes: int = param.Integer(
        default=1, doc="The number of virtual machines that will be allocated for this job in AzureML."
    )
    model: str = param.String(
        doc="The fully or partially qualified name of the model to train/test -e.g. mymodule.configs.MyConfig."
    )
    model_variant: str = param.String(
        default="",
        doc="The name of the model variant to choose, by calling the 'set_model_variant' method. Model variants "
        "are selected before applying commandline overrides of parameters",
    )
    mount_in_azureml: bool = param.Boolean(
        False,
        doc="If False (default), consume datasets in AzureML by downloading at "
        "job start. If True, datasets in AzureML are mounted (read on demand "
        "over the network). When running outside AzureML, datasets will "
        "always be mounted.",
    )
    docker_shm_size: str = param.String("400g", doc="The shared memory in the Docker image for the AzureML VMs.")
    wait_for_completion: bool = param.Boolean(
        default=False,
        doc="If True, wait for AML Run to complete before proceeding. If False, submit the run to AML and exit",
    )
    conda_env: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="The Conda environment file that should be used when submitting the present run to "
        "AzureML. If not specified, the environment file in the current folder or one of its "
        "parents will be used.",
    )
    debug_ddp: DebugDDPOptions = param.ClassSelector(
        default=DebugDDPOptions.OFF,
        class_=DebugDDPOptions,
        doc=f"Flag to override the environment var {DEBUG_DDP_ENV_VAR}"
        "that can be used to trigger logging and collective "
        "synchronization checks to ensure all ranks are synchronized "
        "appropriately. Default is `OFF`. It can be set to either "
        "`INFO` or `DETAIL` for different levels of logging. "
        "`DETAIL` may impact the application performance and thus "
        "should only be used when debugging issues",
    )
    strictly_aml_v1: bool = param.Boolean(
        default=False, doc="If True, use AzureML v1 SDK. If False (default), use the v2 of the SDK"
    )
    identity_based_auth: bool = param.Boolean(
        default=False,
        doc="If True, use identity based authentication to access blob storage data via datastores. If False (default),"
        " then use SAS tokens / account keys if available. Only supported when using v2 sdk.",
    )
    workspace_config_path: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="The path to the AzureML workspace configuration file. If not specified, the "
        "configuration file in the current folder or one of its parents will be used.",
    )
    max_run_duration: str = param.String(
        default="",
        doc="The maximum runtime that is allowed for this job in AzureML. This is given as a floating"
        "point number with a string suffix s, m, h, d for seconds, minutes, hours, day. Examples: '3.5h', '2d'",
    )
    mode: RunnerMode = param.ClassSelector(
        class_=RunnerMode,
        default=RunnerMode.TRAIN,
        doc=f"The mode to run the experiment in. Can be one of '{RunnerMode.TRAIN}' (training and evaluation on the "
        f"test set), or '{RunnerMode.EVAL_FULL}' for evaluation on the full dataset specified by the "
        "'get_eval_data_module' method of the container.",
    )
    log_level: Optional[RunnerMode] = param.ClassSelector(
        class_=LogLevel,
        default=None,
        doc=f"The log level to use. Can be one of {list(map(str, LogLevel))}",
    )

    @property
    def submit_to_azure_ml(self) -> bool:
        """Returns True if the experiment should be submitted to AzureML, False if it should be run locally.

        :return: True if the experiment should be submitted to AzureML, False if it should be run locally."""
        return self.cluster != ""



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from typing import Iterator

from torch.nn.parameter import Parameter
from torch.optim import Adam, Optimizer, SGD
from torch.optim.rmsprop import RMSprop

from health_ml.deep_learning_config import OptimizerParams, OptimizerType


def create_optimizer(config: OptimizerParams, parameters: Iterator[Parameter]) -> Optimizer:
    # Select optimizer type
    if config.optimizer_type in [OptimizerType.Adam, OptimizerType.AMSGrad]:
        return Adam(
            parameters,
            config.l_rate,
            config.adam_betas,
            config.opt_eps,
            config.weight_decay,
            amsgrad=config.optimizer_type == OptimizerType.AMSGrad,
        )
    elif config.optimizer_type == OptimizerType.SGD:
        return SGD(parameters, config.l_rate, config.momentum, weight_decay=config.weight_decay)
    elif config.optimizer_type == OptimizerType.RMSprop:
        return RMSprop(
            parameters, config.l_rate, config.rms_alpha, config.opt_eps, config.weight_decay, config.momentum
        )
    else:
        raise NotImplementedError(f"Optimizer type {config.optimizer_type.value} is not implemented")



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import argparse
from io import TextIOWrapper
import logging
import math
import numbers
import operator
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Mapping, Optional, Union

import torch
from azureml.core import Run, Workspace
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import ProgressBarBase
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.utilities.distributed import rank_zero_only
from pytorch_lightning.utilities.logger import _convert_params, _flatten_dict, _sanitize_params

from health_azure import is_running_in_azure_ml
from health_azure.utils import PathOrString, RUN_CONTEXT, create_aml_run_object


class AzureMLLogger(LightningLoggerBase):
    """
    A Pytorch Lightning logger that stores metrics in the current AzureML run. This logger will always write metrics
    to AzureML if the training run is executed in AzureML. It can optionally also write to AzureML if the training
    run is executed somewhere else, for example on a VM outside of AzureML.
    """

    HYPERPARAMS_NAME = "hyperparams"
    """
    The name under which hyperparameters are written to the AzureML run.
    """

    def __init__(
        self,
        enable_logging_outside_azure_ml: Optional[bool] = False,
        experiment_name: str = "azureml_logger",
        run: Optional[Run] = None,
        run_name: Optional[str] = None,
        workspace: Optional[Workspace] = None,
        workspace_config_path: Optional[Path] = None,
        snapshot_directory: Optional[PathOrString] = None,
    ) -> None:
        """
        :param enable_logging_outside_azure_ml: If True, the AzureML logger will write metrics to AzureML even if
            executed outside of an AzureML run (for example, when working on a separate virtual machine). If False,
            the logger will only write metrics to AzureML if the code is actually running inside of AzureML. Default
            False, do not log outside of AzureML.
        :param experiment_name: The AzureML experiment that should hold the run when executed outside of AzureML.
        :param run: The AzureML run to log to when the ``enable_logging_outside_azure_ml`` flag is True. If None,
            a new run will be created. When finished, the run should be completed by calling ``run.complete()``. The
            logger itself only calls ``run.flush()`` in its ``finalize()`` method.
        :param run_name: An optional name for the run (this will be used as the display name in the AzureML UI). This
            argument only matters when running outside of AzureML.
        :param workspace: If provided, use this workspace to create the run in.
        :param workspace_config_path: Use this path to read workspace configuration json file. If not provided,
            use the workspace specified by the `config.json` file in the current working directory or its parents.
        :param snapshot_directory: The folder that should be included as the code snapshot. By default, no snapshot
            is created. Set this to the folder that contains all the code your experiment uses. You can use a file
            .amlignore to skip specific files or folders, akin to .gitignore..
        """
        super().__init__()
        self.is_running_in_azure_ml = is_running_in_azure_ml()
        self.run: Optional[Run] = None
        self.has_user_provided_run = False
        self.enable_logging_outside_azure_ml = enable_logging_outside_azure_ml
        if self.is_running_in_azure_ml:
            self.run = RUN_CONTEXT
        elif enable_logging_outside_azure_ml:
            if run is not None:
                self.run = run
                self.has_user_provided_run = True
            else:
                try:
                    self.run = create_aml_run_object(
                        experiment_name=experiment_name,
                        run_name=run_name,
                        workspace=workspace,
                        workspace_config_path=workspace_config_path,
                        snapshot_directory=snapshot_directory,
                    )
                    # Display name should already be set when creating the run object, but this does not happen.
                    # In unit tests, the run has the expected display name, but not here. Hence, set it again.
                    self.run.display_name = run_name
                except Exception as ex:
                    logging.error(f"Unable to create an AzureML run to store the results because of {ex}.")
                    raise
            print(f"Writing metrics to run {self.run.id} in experiment {self.run.experiment.name}.")
            print(f"To check progress, visit this URL: {self.run.get_portal_url()}")
        else:
            print(
                "AzureMLLogger will not write any logs because it is running outside AzureML, and the "
                "'enable_logging_outside_azure_ml' flag is set to False"
            )

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        """
        Writes the given metrics dictionary to the AzureML run context. If the metrics dictionary has an `epoch` key,
        the `step` value (x-axis for plots) is left empty. If there is no `epoch` key, the `step` value is taken
        from the function argument. This is the case for metrics that are logged with the `on_step=True` flag.

        :param metrics: A dictionary with metrics to log. Keys are strings, values are floating point numbers.
        :param step: The trainer global step for logging.
        """
        logging.debug(f"AzureMLLogger step={step}: {metrics}")
        if self.run is None:
            return
        is_epoch_metric = "epoch" in metrics
        for key, value in metrics.items():
            # Log all epoch-level metrics without the step information
            # All step-level metrics with step
            self.run.log(key, value, step=None if is_epoch_metric else step)

    @rank_zero_only
    def log_hyperparams(self, params: Union[argparse.Namespace, Dict[str, Any]]) -> None:
        """
        Logs the given model hyperparameters to AzureML as a table. Namespaces are converted to dictionaries.
        Nested dictionaries are flattened out. The hyperparameters are then written as a table with two columns
        "name" and "value".
        """
        if self.run is None:
            return
        if params is None:
            return
        params_final = _preprocess_hyperparams(params)
        if len(params_final) > 0:
            # Log hyperparameters as a table with 2 columns. Each "step" is one hyperparameter
            self.run.log_table(
                self.HYPERPARAMS_NAME, {"name": list(params_final.keys()), "value": list(params_final.values())}
            )

    def experiment(self) -> Any:
        return None

    def name(self) -> Any:
        return ""

    def version(self) -> int:
        return 0

    def finalize(self, status: str) -> None:
        if self.enable_logging_outside_azure_ml and not self.is_running_in_azure_ml and self.run is not None:
            if self.has_user_provided_run:
                # The logger uses a run that was provided by the user: Flush it, but do not complete it.
                # The user should complete the run after finishing the experiment. This is important when running
                # training outside of AzureML, so that training and inference metrics can be written to the same run.
                self.run.flush()
            else:
                # Run.complete should only be called if we created an AzureML run here in the constructor.
                self.run.complete()


class AzureMLProgressBar(ProgressBarBase):
    """
    A PL progress bar that works better in AzureML. It prints timestamps for each message, and works well with a setup
    where there is no direct access to the console.

    Usage example:
        >>> from health_ml.utils import AzureMLProgressBar
        >>> from pytorch_lightning import Trainer
        >>> progress = AzureMLProgressBar(refresh_rate=100)
        >>> trainer = Trainer(callbacks=[progress])
    """

    PROGRESS_STAGE_TRAIN = "Training"
    """A string that indicates that the trainer loop is presently in training mode."""
    PROGRESS_STAGE_VAL = "Validation"
    """A string that indicates that the trainer loop is presently in validation mode."""
    PROGRESS_STAGE_TEST = "Testing"
    """A string that indicates that the trainer loop is presently in testing mode."""
    PROGRESS_STAGE_PREDICT = "Prediction"
    """A string that indicates that the trainer loop is presently in prediction mode."""

    def __init__(self, refresh_rate: int = 50, print_timestamp: bool = True, write_to_logging_info: bool = False):
        """
        :param refresh_rate: The number of steps after which the progress should be printed out.
        :param print_timestamp: If True, each message that the progress bar prints will be prefixed with the current
            time in UTC. If False, no such prefix will be added.
        :param write_to_logging_info: If True, the progress information will be printed via logging.info. If False,
            it will be printed to stdout via print.
        """
        super().__init__()
        self._refresh_rate = refresh_rate
        self._enabled = True
        self.stage = ""
        self.stage_start_time = 0.0
        self.total_num_batches = 0.0
        self.write_to_logging_info = write_to_logging_info
        self.print_timestamp = print_timestamp

    @property
    def refresh_rate(self) -> int:
        return self._refresh_rate

    @property
    def is_enabled(self) -> bool:
        return self._enabled and self.refresh_rate > 0

    @property
    def is_disabled(self) -> bool:
        return not self.is_enabled

    @property
    def total_test_batches(self) -> int:
        assert self._trainer is not None
        return sum(self.trainer.num_test_batches)  # type: ignore

    @property
    def total_predict_batches(self) -> int:
        assert self._trainer is not None
        return sum(self.trainer.num_predict_batches)  # type: ignore

    def disable(self) -> None:
        self._enabled = False

    def enable(self) -> None:
        self._enabled = True

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_epoch_start(trainer, pl_module)
        self.start_stage(self.PROGRESS_STAGE_TRAIN, self.total_train_batches)

    def on_validation_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_validation_start(trainer, pl_module)
        self.start_stage(self.PROGRESS_STAGE_VAL, self.total_val_batches)

    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_test_epoch_start(trainer, pl_module)
        self.start_stage(self.PROGRESS_STAGE_TEST, self.total_test_batches)

    def on_predict_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_predict_epoch_start(trainer, pl_module)
        self.start_stage(self.PROGRESS_STAGE_PREDICT, self.total_predict_batches)

    def start_stage(self, stage: str, total_num_batches: Union[int, float]) -> None:
        """
        Sets the information that a new stage of the PL loop is starting. The stage will be available in
        self.stage, total_num_batches in self.total_num_batches. The time when this method was called is recorded in
        self.stage_start_time

        :param stage: The string name of the stage that has just started.
        :param total_num_batches: The total number of batches that need to be processed in this stage. This is used
            only for progress reporting.
        """
        self.stage = stage
        self.total_num_batches = total_num_batches
        self.stage_start_time = time.time()

    def on_train_batch_end(self, *args: Any, **kwargs: Any) -> None:
        super().on_train_batch_end(*args, **kwargs)
        self.update_progress(batches_processed=self.train_batch_idx)

    def on_validation_batch_end(self, *args: Any, **kwargs: Any) -> None:
        super().on_validation_batch_end(*args, **kwargs)
        self.update_progress(batches_processed=self.val_batch_idx)

    def on_test_batch_end(self, *args: Any, **kwargs: Any) -> None:
        super().on_test_batch_end(*args, **kwargs)
        self.update_progress(batches_processed=self.test_batch_idx)

    def on_predict_batch_end(self, *args: Any, **kwargs: Any) -> None:
        super().on_predict_batch_end(*args, **kwargs)
        self.update_progress(batches_processed=self.predict_batch_idx)

    def update_progress(self, batches_processed: int) -> None:
        """
        Writes progress information once the refresh interval is full.

        :param batches_processed: The number of batches that have been processed for the current stage.
        """

        def to_minutes(time_sec: float) -> str:
            minutes = int(time_sec / 60)
            seconds = int(time_sec % 60)
            return f"{minutes:02}:{seconds:02}"

        should_update = self.is_enabled and (
            batches_processed % self.refresh_rate == 0 or batches_processed == self.total_num_batches
        )  # noqa: W503
        if not should_update:
            return
        prefix = f"{self.stage}"
        assert self.trainer is not None and self.trainer.lightning_module is not None  # for pyright
        if self.stage in [self.PROGRESS_STAGE_TRAIN, self.PROGRESS_STAGE_VAL]:
            prefix += f" epoch {self.trainer.current_epoch}"
        if self.stage == self.PROGRESS_STAGE_TRAIN:
            prefix += f" (step {self.trainer.lightning_module.global_step})"
        prefix += ": "
        time_elapsed = time.time() - self.stage_start_time
        time_elapsed_min = to_minutes(time_elapsed)
        if math.isinf(self.total_num_batches):
            # Can't print out per-cent progress or time estimates if the data is infinite
            message = f"{prefix}{batches_processed:4} batches completed, {time_elapsed_min} since epoch start"
        else:
            fraction_completed = batches_processed / self.total_num_batches
            percent_completed = int(fraction_completed * 100)
            estimated_epoch_duration = time_elapsed / fraction_completed
            message = (
                f"{prefix}{batches_processed:4}/{self.total_num_batches} ({percent_completed:3}%) completed. "
                f"{time_elapsed_min} since epoch start, estimated total epoch time ~ "
                f"{to_minutes(estimated_epoch_duration)}"
            )
        self._print(message)

    def _print(self, message: str) -> None:
        if self.print_timestamp:
            timestamp = datetime.utcnow().strftime("%Y-%m-%dT%H%M%SZ ")
            message = timestamp + message
        if self.write_to_logging_info:
            logging.info(message)
        else:
            print(message)
            sys.stdout.flush()


def _preprocess_hyperparams(params: Any) -> Dict[str, str]:
    """
    Converts arbitrary hyperparameters to a simple dictionary structure, in particular argparse Namespaces.
    Nested dictionaries are converted to folder-like strings, like ``{'a': {'b': 'c'}} -> {'a/b': 'c'}``.
    All hyperparameter values are converted to strings, because Run.log_table can't deal with mixed datatypes.
    :param params: The parameters to convert
    :return: A dictionary mapping from string to string.
    """
    # Convert from Namespace to dictionary
    params = _convert_params(params)
    # Convert nested dictionaries to folder-like structure
    params = _flatten_dict(params)
    # Convert anything that is not a primitive type to str
    params_final = _sanitize_params(params)
    if not isinstance(params_final, dict):
        raise ValueError(f"Expected the converted hyperparameters to be a dictionary, but got {type(params)}")
    return {str(key): str(value) for key, value in params_final.items()}


def log_on_epoch(
    module: LightningModule,
    name: Optional[str] = None,
    value: Optional[Any] = None,
    metrics: Optional[Mapping[str, Any]] = None,
    reduce_fx: Callable = torch.mean,
    sync_dist: Optional[bool] = None,
) -> None:
    """
    Write a dictionary with metrics and/or an individual metric as a name/value pair to the loggers of the given module.
    Metrics are always logged upon epoch completion.
    The metrics in question first synchronized across GPUs if DDP with >1 node is used,. Afterwards, they are
    aggregated across all steps via the reduce_fx (default: mean).
    Metrics that are fed in as plain numbers rather than tensors (for example, plain Python integers) are converted
    to tensors before logging, to enable synchronization across GPUs if needed.

    :param name: The name of the metric to log.
    :param value: The actual value of the metric to log.
    :param metrics: A dictionary with metrics to log.
    :param module: The PyTorch Lightning module where the metrics should be logged.
    :param sync_dist: If not None, use this value for the sync_dist argument to module.log. If None,
        set it automatically depending on the use of DDP. Set this to False if you want to log metrics that are only
        available on Rank 0 of a DDP job.
    :param reduce_fx: The reduce function to apply to the per-step values, after synchronizing the tensors across GPUs.
        Default: torch.mean
    """
    assert module.trainer is not None, "No trainer is set for this module."
    if operator.xor(name is None, value is None):
        raise ValueError("Both or neither of 'name' and 'value' must be provided.")
    is_sync_dist = module.trainer.world_size > 1 if sync_dist is None else sync_dist
    metrics = metrics or {}
    if name is not None:
        metrics[name] = value  # type: ignore
    metrics_as_tensors = {
        key: torch.tensor(value, dtype=torch.float, device=module.device)
        if isinstance(value, numbers.Number)
        else value
        for key, value in metrics.items()
    }
    module.log_dict(metrics_as_tensors, on_epoch=True, on_step=False, sync_dist=is_sync_dist, reduce_fx=reduce_fx)


def log_learning_rate(module: LightningModule, name: str = "learning_rate") -> None:
    """
    Logs the learning rate(s) used by the given module. Multiple learning rate schedulers and/or multiple rates per
    scheduler are supported. The learning rates are logged under the given name. If multiple scheduler and/or multiple
    rates are used, a suffix like "/0/1" is added, to indicate the learning rate for scheduler 0, index 1, for example.
    Learning rates are logged on epoch.

    :param module: The module for which the learning rates should be logged.
    :param name: The name to use when logging the learning rates.
    """
    schedulers = module.lr_schedulers()
    if schedulers is None:
        raise ValueError("Learning rate logging can only be used during training.")
    single_scheduler = not isinstance(schedulers, list)
    if single_scheduler:
        schedulers = [schedulers]  # type: ignore
    lr_0 = schedulers[0].get_last_lr()  # type: ignore
    singleton_lr = single_scheduler and len(lr_0) == 1
    logged = {}
    for i, scheduler in enumerate(schedulers):  # type: ignore
        for j, lr_j in enumerate(scheduler.get_last_lr()):  # type: ignore
            full_name = name if singleton_lr else f"{name}/{i}/{j}"
            logged[full_name] = lr_j
    log_on_epoch(module, metrics=logged)


class ConsoleAndFileOutput(TextIOWrapper):
    """A file-like object that writes to both the console (sys.stdout) and an output file.
    The caller needs to ensure that the file is closed properly after using this object.

    :param TextIOWrapper: The file to write to.
    """

    def __init__(self, file: TextIOWrapper) -> None:
        self.file = file
        self.sys_stdout = sys.stdout

    def write(self, __buffer: Any) -> int:
        """Writes the buffer to both the file and to the console.

        :param __buffer: The buffer to write.
        :return: The number of bytes written to the console
        """
        self.file.write(__buffer)
        return self.sys_stdout.write(__buffer)

    def flush(self) -> None:
        """Flush both the output file and the console."""
        self.file.flush()
        self.sys_stdout.flush()

    def isatty(self) -> bool:
        """Returns if the given file is a terminal. Since the class is writing to a file, this always returns False."""
        return False

    def read(self, size: int = -1) -> str:  # type: ignore
        raise NotImplementedError("Read is not supported")

    def readline(self, size: int = -1) -> str:  # type: ignore
        raise NotImplementedError("Readline is not supported")



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
from pathlib import Path
from typing import List, Optional

from azureml.core import Run
from pytorch_lightning import LightningDataModule, Trainer, seed_everything

from health_azure import AzureRunInfo
from health_azure.utils import (
    PARENT_RUN_CONTEXT,
    RUN_CONTEXT,
    create_aml_run_object,
    create_run_recovery_id,
    is_running_in_azure_ml,
)
from health_ml.experiment_config import ExperimentConfig
from health_ml.lightning_container import LightningContainer
from health_ml.model_trainer import create_lightning_trainer
from health_ml.utils import fixed_paths
from health_ml.utils.checkpoint_handler import CheckpointHandler
from health_ml.utils.common_utils import (
    EFFECTIVE_RANDOM_SEED_KEY_NAME,
    RUN_RECOVERY_FROM_ID_KEY_NAME,
    RUN_RECOVERY_ID_KEY,
    change_working_directory,
    seed_monai_if_available,
)
from health_ml.utils.lightning_loggers import StoringLogger, get_mlflow_run_id_from_trainer
from health_ml.utils.type_annotations import PathOrString


def check_dataset_folder_exists(local_dataset: PathOrString) -> Path:
    """
    Checks if a folder with a local dataset exists. If it does exist, return the argument converted
    to a Path instance. If it does not exist, raise a FileNotFoundError.

    :param local_dataset: The dataset folder to check.
    :return: The local_dataset argument, converted to a Path.
    """
    expected_dir = Path(local_dataset)
    if not expected_dir.is_dir():
        raise FileNotFoundError(f"The model uses a dataset in {expected_dir}, but that does not exist.")
    logging.info(f"Model will use the local dataset provided in {expected_dir}")
    return expected_dir


class RunnerBase:
    """
    A base class with operations that are shared between the training/test runner and the evaluation-only runner.
    """

    def __init__(
        self, experiment_config: ExperimentConfig, container: LightningContainer, project_root: Optional[Path] = None
    ) -> None:
        """
        Driver class to run an ML experiment. Note that the project root argument MUST be supplied when using hi-ml
        as a package!

        :param experiment_config: The ExperimentConfig object to use for training.
        :param container: The LightningContainer object to use for training.
        :param project_root: Project root. This should only be omitted if calling run_ml from the test suite. Supplying
            it is crucial when using hi-ml as a package or submodule!
        """
        self.container = container
        self.experiment_config = experiment_config
        self.container.num_nodes = self.experiment_config.num_nodes
        self.container.runner_mode = self.experiment_config.mode
        self.project_root: Path = project_root or fixed_paths.repository_root_directory()
        self.storing_logger: Optional[StoringLogger] = None
        self._has_setup_run = False
        self.checkpoint_handler = CheckpointHandler(
            container=self.container, project_root=self.project_root, run_context=RUN_CONTEXT
        )
        self.trainer: Optional[Trainer] = None
        self.azureml_run_for_logging: Optional[Run] = None
        self.mlflow_run_for_logging: Optional[str] = None
        # This is passed to trainer.validate and trainer.test in inference mode
        self.inference_checkpoint: Optional[str] = None

    def validate(self) -> None:
        """
        Checks if all arguments and settings of the object are correct.
        """
        pass

    def setup_azureml(self) -> None:
        """
        Execute setup steps that are specific to AzureML.
        """
        if PARENT_RUN_CONTEXT is not None:
            # Set metadata for the run in AzureML if running in a Hyperdrive job.
            run_tags_parent = PARENT_RUN_CONTEXT.get_tags()
            tags_to_copy = [
                "tag",
                "model_name",
                "execution_mode",
                "recovered_from",
                "friendly_name",
                "build_number",
                "build_user",
                RUN_RECOVERY_FROM_ID_KEY_NAME,
            ]
            new_tags = {tag: run_tags_parent.get(tag, "") for tag in tags_to_copy}
            new_tags[RUN_RECOVERY_ID_KEY] = create_run_recovery_id(run=RUN_CONTEXT)
            new_tags[EFFECTIVE_RANDOM_SEED_KEY_NAME] = str(self.container.get_effective_random_seed())
            RUN_CONTEXT.set_tags(new_tags)

    def setup(self, azure_run_info: Optional[AzureRunInfo] = None) -> None:
        """
        Sets the random seeds, calls the setup method on the LightningContainer and then creates the actual
        Lightning modules.

        :param azure_run_info: When running in AzureML or on a local VM, this contains the paths to the datasets.
            This can be missing when running in unit tests, where the local dataset paths are already populated.
        """
        if self._has_setup_run:
            return
        if azure_run_info:
            # Set up the paths to the datasets. azure_run_info already has all necessary information, using either
            # the provided local datasets for VM runs, or the AzureML mount points when running in AML.
            # This must happen before container setup because that could already read datasets.
            logging.info("Setting tags from parent run.")
            input_datasets = azure_run_info.input_datasets
            logging.info(f"Setting the following datasets as local datasets: {input_datasets}")
            if len(input_datasets) > 0:
                local_datasets: List[Path] = []
                for i, dataset in enumerate(input_datasets):
                    if dataset is None:
                        raise ValueError(f"Invalid setup: The dataset at index {i} is None")
                    local_datasets.append(check_dataset_folder_exists(dataset))
                self.container.local_datasets = local_datasets  # type: ignore
        # Ensure that we use fixed seeds before initializing the PyTorch models.
        # MONAI needs a separate method to make all transforms deterministic by default
        seed = self.container.get_effective_random_seed()
        seed_monai_if_available(seed)
        seed_everything(seed)

        # Creating the folder structure must happen before the LightningModule is created, because the output
        # parameters of the container will be copied into the module.
        self.container.create_filesystem(self.project_root)

        # configure recovery container if provided
        self.checkpoint_handler.download_recovery_checkpoints_or_weights()

        # Create an AzureML run for logging if running outside AzureML.
        self.create_logger()

        self.container.setup()
        self.container.create_lightning_module_and_store()
        self._has_setup_run = True

        if is_running_in_azure_ml():
            self.setup_azureml()

    def create_logger(self) -> None:
        """
        Create an AzureML run for logging if running outside AzureML. This run will be used for metrics logging
        during both training and inference. We can't rely on the automatically generated run inside the AzureMLLogger
        class because two of those logger objects will be created, so training and inference metrics would be logged
        in different runs.
        """
        if self.container.log_from_vm:
            run = create_aml_run_object(experiment_name=self.container.effective_experiment_name)
            # Display name should already be set when creating the Run object, but in some scenarios this
            # does not happen. Hence, set it again.
            run.display_name = self.container.tag if self.container.tag else None
            self.azureml_run_for_logging = run

    def get_data_module(self) -> LightningDataModule:
        """
        Reads the datamodule that should be used for training or valuation from the container. This must be
        overridden in subclasses.
        """
        raise NotImplementedError()

    def set_trainer_for_inference(self) -> None:
        """Set the runner's PL Trainer object that should be used when running inference on the validation or test set.
        We run inference on a single device because distributed strategies such as DDP use DistributedSampler
        internally, which replicates some samples to make sure all devices have the same batch size in case of
        uneven inputs which biases the results."""
        mlflow_run_id = get_mlflow_run_id_from_trainer(self.trainer)
        self.container.max_num_gpus = self.container.max_num_gpus_inference
        self.trainer, _ = create_lightning_trainer(
            container=self.container,
            num_nodes=1,
            azureml_run_for_logging=self.azureml_run_for_logging,
            mlflow_run_for_logging=mlflow_run_id,
        )

    def init_inference(self) -> None:
        """Prepare the runner for inference on validation set, test set, or a full dataset.
        The following steps are performed:

        1. Get the checkpoint to use for inference. This is either the checkpoint from the last training epoch or the
        one specified in src_checkpoint argument.

        2. Create a new trainer instance for inference. This is necessary because the trainer is created with a single
        device in contrast to training that uses DDP if multiple GPUs are available.

        3. Create a new data module instance for inference to account for any requested changes in the dataloading
        parameters (e.g. batch_size, max_num_workers, etc) as part of on_run_extra_validation_epoch.
        """
        logging.info("Preparing runner for inference.")
        self.inference_checkpoint = str(self.checkpoint_handler.get_checkpoint_to_test())
        self.set_trainer_for_inference()
        self.data_module = self.get_data_module()

    def run_inference(self) -> None:
        """Run inference on the test set for all models. This is done by calling the LightningModule.test_step method.
        If the LightningModule.test_step method is not overridden, then this method does nothing. The cwd is changed to
        the outputs folder so that the model can write to current working directory, and still everything is put into
        the right place in AzureML (there, only the contents of the "outputs" folder is treated as a result file).
        """
        if self.container.has_custom_test_step():
            logging.info("Running inference via the LightningModule.test_step method")
            with change_working_directory(self.container.outputs_folder):
                assert self.trainer, "Trainer should be initialized before inference. Call self.init_inference()."
                _ = self.trainer.test(
                    self.container.model, datamodule=self.data_module, ckpt_path=self.inference_checkpoint
                )
        else:
            logging.warning("None of the suitable test methods is overridden. Skipping inference completely.")

    def run(self) -> None:
        """
        Run the training or evaluation. This method must be overridden in subclasses.
        """
        pass

    def run_and_cleanup(self, azure_run_info: Optional[AzureRunInfo] = None) -> None:
        """
        Run the training or evaluation via `self.run` and cleanup afterwards.

        :param azure_run_info: When running in AzureML or on a local VM, this contains the paths to the datasets.
            This can be missing when running in unit tests, where the local dataset paths are already populated.
        """
        self.setup(azure_run_info)
        try:
            self.run()
        finally:
            if self.azureml_run_for_logging is not None:
                try:
                    self.azureml_run_for_logging.complete()
                except Exception as ex:
                    logging.error("Failed to complete AzureML run: %s", ex)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from dataclasses import dataclass
from pathlib import Path
from health_azure.paths import git_repo_root_folder, is_himl_used_from_git_repo


def repository_root_directory() -> Path:
    """
    Gets the full path to the root directory that holds the present repository.
    This function should only be called if the repository is available (for example, in unit tests).
    It will raise a ValueError if the repo is not available.

    :param path: if provided, a relative path to append to the absolute path to the repository root.
    :return: The full path to the repository's root directory, with symlinks resolved if any.
    """
    if is_himl_used_from_git_repo():
        return git_repo_root_folder()
    raise ValueError("This function should not be used if hi-ml is used as an installed package.")


@dataclass(frozen=True)
class OutputFolderForTests:
    """
    Data class for the output directories for a given test
    """

    root_dir: Path

    def create_file_or_folder_path(self, file_or_folder_name: str) -> Path:
        """
        Creates a full path for the given file or folder name relative to the root directory stored in the present
        object.
        :param file_or_folder_name: Name of file or folder to be created under root_dir
        """
        return self.root_dir / file_or_folder_name

    def make_sub_dir(self, dir_name: str) -> Path:
        """
        Makes a sub directory under root_dir
        :param dir_name: Name of subdirectory to be created.
        """
        sub_dir_path = self.create_file_or_folder_path(dir_name)
        sub_dir_path.mkdir()
        return sub_dir_path



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
import matplotlib
import os

from health_azure import health_azure_package_setup, set_logging_levels


def health_ml_package_setup() -> None:
    """
    Set up the Python packages where needed. In particular, reduce the logging level for some of the used
    libraries, which are particularly talkative in DEBUG mode. Usually when running in DEBUG mode, we want
    diagnostics about the model building itself, but not for the underlying libraries.

    It also adds workarounds for known issues in some packages.
    """
    health_azure_package_setup()
    module_levels = {
        # DEBUG level info when opening checkpoint and other files
        'fsspec': logging.INFO,
        # Matplotlib is also very talkative in DEBUG mode, filling half of the log file in a PR build.
        'matplotlib': logging.INFO,
        # Jupyter notebook report generation
        'nbconvert': logging.INFO,
        # Numba code generation is extremely talkative in DEBUG mode, disable that.
        'numba': logging.WARNING,
        # PIL prints out byte-level information when loading PNG files in DEBUG mode
        "PIL": logging.INFO,
        # Jupyter notebook report generation
        'papermill': logging.INFO,
    }
    set_logging_levels(module_levels)
    # This is working around a spurious error message thrown by MKL, see
    # https://github.com/pytorch/pytorch/issues/37377
    os.environ['MKL_THREADING_LAYER'] = 'GNU'
    # Workaround for issues with matplotlib on some X servers, see
    # https://stackoverflow.com/questions/45993879/matplot-lib-fatal-io-error-25-inappropriate-ioctl-for-device-on-x
    # -server-loc
    matplotlib.use('Agg')



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import base64
import itertools
import mimetypes
from datetime import datetime
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional, OrderedDict, Tuple

import jinja2
from ruamel.yaml import YAML
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from health_azure.utils import (
    download_files_from_run_id,
    get_aml_run_from_run_id,
    download_files_from_hyperdrive_children,
)

CLOSE_DOC_TAGS = "</p>\n</div>\n</body>\n</html>"
IMAGE_KEY_HTML = "IMAGEPATHSHTML"
TABLE_KEY_HTML = "TABLEKEYHTML"
REPORT_CONTENTS_KEY = "report_contents"
DEFAULT_OUTPUTS_FOLDER = "outputs"
DEFAULT_FIGSIZE = (15, 15)
DEFAULT_NUM_COLS = 2


class ReportComponentKey(Enum):
    IMAGE = "image"
    IMAGE_GALLERY = "image_gallery"
    TABLE = "table"
    TEXT = "text"
    TYPE = "type"
    VALUE = "value"


class HTMLReport:
    """
    Create an HTML report which Azure ML is able to render. If you do not provide a title or output folder,
    this will generate a report at "outputs/report.html"
    """

    def __init__(self, title: str = "Report", output_folder: str = DEFAULT_OUTPUTS_FOLDER):
        self.report_title = title
        self.output_folder = output_folder

        report_folder = Path(output_folder)
        report_folder.mkdir(exist_ok=True, parents=True)

        self.report_folder = report_folder
        self.report_path_html = (report_folder / title.lower().replace(" ", "_")).with_suffix('.html')
        self.report_html = ""
        self.template = ""
        self.template_path = self._create_template()

        self.env = jinja2.Environment(loader=jinja2.FileSystemLoader('/'))
        self.render_kwargs: Dict[str, Any] = {"title": title}

    def validate(self) -> None:
        """
        For our definition, the rendered HTML must contain exactly one open and closing tags doctype, head and body
        If any of these are missing from the repport_html, we raise a ValueError

        :return:
        """
        # If this function is called before the report is rendered, self.html_report will be empty.
        # calling render() will populate this attribute
        if len(self.report_html) == 0:
            self.render(save_html=False)
        expected_tags = ["<!DOCTYPE html>", "<head>", "</head>", "<body>", "</body>", "</html>"]
        for tag in expected_tags:
            if self.report_html.count(tag) < 1:
                raise ValueError(f"report_html is missing the tag {tag}. This will cause problems with rendering")
            elif self.report_html.count(tag) > 1:
                raise ValueError(
                    f"report_html contains more than one tag {tag}. This will cause problems with rendering"
                )

    @staticmethod
    def _remove_html_end(report_stream: str) -> str:
        """
        Before adding additional components to the HTML report, we must remove the closing tags

        :param report_stream: A string representing the content of the report thus far
        :return: A modified string, without closing tags
        """
        return report_stream.replace(CLOSE_DOC_TAGS, "")

    def _add_html_end(self) -> None:
        """
        Once we have added components to our template, this method is called, to add back the HTML
        closing tags. validate is called to check the correct number of closing tags exist.

        :return: A modified string, with closing tags
        """
        self.template += f"{CLOSE_DOC_TAGS}"
        self.validate()

    def _create_template(self) -> Path:
        template_path = self.report_folder / "template.html"
        template_path.touch(exist_ok=True)

        self.template += (
            """<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="UTF-8">
<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet">
<title> {{title}} </title>
</head>
<body>
<div class="container-fluid">
<div class="container">
<h1> {{title}} </h1>
</div>
<p>
"""
            + CLOSE_DOC_TAGS
        )

        return template_path

    def add_to_template(self, template_addition: str) -> None:
        """
        Performs the three necessary tasks to add a section to the template. Firstly, removes existing closing
        tags. Next, adds the content (passed as a string to this method). Finally, adds back the closing tags.

        :param template_addition: The string to be added to the template
        """
        self.template = self._remove_html_end(self.template)
        self.template += template_addition
        self._add_html_end()

    def add_heading(self, text: str, level: int = 2, tag_class: str = '') -> None:
        """
        Add a heading to the report content. If you wish to provide your own css classes, you can
        tag this heading by providing the class name in the "tag_class" parameter.

        :param text: The contents of the heading to add to the report
        :param level: The heading level, e.g. 2 for `<h2>` etc.
        :param tag_class: An optional class name to apply styling to the text
        """
        if level < 1 or level > 5:
            raise ValueError(f"Level must be an integer between 1 and 5 (inclusive), but got {level}")
        class_spec = f" class={tag_class}" if tag_class else ""
        template_addition = f"""<div class="container" >
        <h{level}{class_spec}>{text}</h{level}>
        </div>"""
        self.add_to_template(template_addition)

    def add_text(self, text: str, tag_class: str = '') -> None:
        """
        Add text to the report content, in the form of a new paragraph. This will start on a new line by default.
        If you wish to provide your own css classes, you can tag this paragraph by providing the class name in the
        "tag_class" parameter.

        :param text: The text to add to the report
        :param tag_class: An optional class name to apply styling to the text
        """
        p_tag_open = f"<p class={tag_class}>" if tag_class is not None else "<p>"

        template_addition = f"""<div class="container" >
        {p_tag_open}{text}
        </div>
        <br>"""
        self.add_to_template(template_addition)

    def _add_tables_to_report(self, tables: List[pd.DataFrame]) -> None:
        """
        Add one or more tables (in the form of Pandas DataFrames) to the report.

        :param tables: A list of one or more Pandas DataFrame to be rendered in the report
        """
        for table in tables:
            num_existing_tables = self.template.count("table.to_html")
            table_key = f"{TABLE_KEY_HTML}_{num_existing_tables}"  # starts at zero

            template_addition = (
                """<div class="container" >
            {% for table in """
                + table_key
                + """ %}
                {{ table.to_html(classes=[ "table"], justify="center") | safe }}
            {% endfor %}
            </div>
            <br>"""
            )
            self.add_to_template(template_addition)

            self.render_kwargs.update({table_key: [table]})

    def add_tables(
        self, tables: Optional[List[pd.DataFrame]] = None, table_paths_or_dir: Optional[List[Path]] = None
    ) -> None:
        """
        Add one or more tables to your report. The table can either be passed as a Pandas DataFrame object, or
        a list of path to one or more .csv files, or a directory of csv files containing your tables.
        If neither of these parameters are provided, an Exception will be raised.

        :param tables: An optional list of one or more Pandas DataFrames to be rendered in the report
        :param table_paths_or_dir: An optional list of one or more paths to .csv files containing the tables
            to be rendered on the report
        :raises ValueError: If neither a list of tables nor a list of paths is provided
        """
        if tables is None and table_paths_or_dir is None:
            raise ValueError("One of tables or table_paths_or_dir must be provided")

        tables = tables or []
        if table_paths_or_dir is not None:
            for table_path_or_dir in table_paths_or_dir:
                table_path_or_dir = Path(table_path_or_dir)
                if table_path_or_dir.is_dir():
                    for table_path in table_path_or_dir.iterdir():
                        self.add_tables(table_paths_or_dir=[table_path])
                    return
                else:
                    tables.append(pd.read_csv(table_path_or_dir))

        assert len(tables) > 0, "No tables were found"
        self._add_tables_to_report(tables)

    def _add_image_to_report(self, image_path: Path, base64_encode: bool = False) -> None:
        """
        Given a path to an image, add it to the report template and to the report arguments
        for rendering later. Optionally encode as base64 - this is useful if a standalone
        report is required but leads to a much larger report file size, so is False by default

        :param image_path: The paths to the image to be added
        :param base64_encode: If True, encode the image as base64 in the HTML report. Default is False
        """
        if self.report_folder in image_path.parents:
            img_path_html = image_path.relative_to(self.report_folder)
        else:
            img_path_html = image_path

        image_key_html = IMAGE_KEY_HTML + "_0"
        # Increment the image name so as not to replace other images
        num_existing_images = self.template.count("<img src=")

        image_key_html = image_key_html.split("_")[0] + f"_{num_existing_images}"

        template_addition = (
            """<div class="container">
        {% for image_path in """
            + image_key_html
            + """ %}
            <img src={{image_path}} alt={{image_path}}>
        {% endfor %}
        </div>
        <br>"""
        )
        self.add_to_template(template_addition)

        # Add these keys and paths to the keyword args for rendering later
        if base64_encode:
            with open(image_path, "rb") as f_path:
                img_data = f_path.read()
                img_data_base64_bytes = base64.b64encode(img_data)
                img_data_base64_str = img_data_base64_bytes.decode()

            img_type: str = mimetypes.guess_type(str(img_path_html))[0]  # type: ignore
            img_path_str = "data:" + img_type + ";base64," + img_data_base64_str
        else:
            img_path_str = str(img_path_html)
        self.render_kwargs.update({image_key_html: [img_path_str]})

    def add_images(self, image_paths_or_dir: List[Path], base64_encode: bool = False) -> None:
        """
        Given a path to one or more image files, or a directory containing image files, embeds the image on the
        report. If the path is within the report folder, the relative path will be used.
        This is to ensure that the HTML document is able to locate and embed the image.
        Otherwise, the image path is not altered.

        :param image_paths_or_dir: The paths to the image(s), or a directory containing images to be embedded
        :param base64_encode: If True, encode image data as base64 in the HTML report. Default is False
        """
        if len(image_paths_or_dir) == 0:
            raise ValueError("add_image expects a list of image_paths")

        for image_path_or_dir in image_paths_or_dir:
            image_path_or_dir = Path(image_path_or_dir)
            if image_path_or_dir.is_dir():
                for image_path in image_path_or_dir.iterdir():
                    self.add_images([image_path], base64_encode=base64_encode)
            else:
                self._add_image_to_report(image_path_or_dir, base64_encode=base64_encode)

    @classmethod
    def load_imgs_onto_subplot(
        cls, img_folder_or_paths: List[Path], num_plot_columns: int = 2, figsize: Tuple[int, int] = DEFAULT_FIGSIZE
    ) -> plt.Figure:
        """
        Given a list of one or more paths, either to a folder containing multiple images, or multiple image
        paths, loads each of the images adds to a single chart

        :param img_folder_or_paths: A list containing either a single path to a folder containing all of the images to
            add to the gallery, or multiple paths to specific image files
        :param num_plot_columns: The number of columns of images to plot, defaults to 2
        :param figsize: The size of the overall figure
        :return: A matplotlib Figure object
        """
        if num_plot_columns <= 1:
            raise ValueError("Can't have less than one column in your plot")

        # if a single path is provided, we take its contents as the plot paths. If a list of image paths is provided
        # we use that.
        plot_paths: List[Path] = []
        for folder_or_path in img_folder_or_paths:
            if folder_or_path.is_dir():
                # Note: this will add every file within the named folder. Make sure these are only image files.
                plot_paths.extend(list(chain(list(folder_or_path.rglob("*")))))
            else:
                plot_paths.append(folder_or_path)

        num_plots = len(plot_paths)
        num_plot_rows = int(np.ceil(num_plots / num_plot_columns))
        fig = plt.figure(figsize=figsize)
        for i, j in itertools.product(range(num_plot_rows), range(num_plot_columns)):
            plot_index = (i * num_plot_columns) + j
            if plot_index >= num_plots:
                break
            plot_path = plot_paths[plot_index]
            with open(plot_path, "rb") as f_path:
                img_arr = plt.imread(f_path)

                axs: plt.Axes = fig.add_subplot(num_plot_rows, num_plot_columns, plot_index + 1)
                axs.set_axis_off()
                plt.imshow(img_arr)

        plt.tight_layout()
        return fig

    def add_image_gallery(
        self,
        image_folder_or_paths: List[Path],
        figsize: Tuple[int, int] = DEFAULT_FIGSIZE,
        num_cols: int = DEFAULT_NUM_COLS,
        base64_encode: bool = False,
    ) -> None:
        """
        Given a list of one or more paths, either to a folder containing multiple images, or multiple image
        paths, loads each of the images adds to a single chart create a "gallery" i.e. a plot containing
        all of these images, and add it to the report

        :param image_folder_or_paths: A list containing either a single path to a folder containing all of the images
            to add to the gallery, or multiple paths to specific image files
        :param figsize: The size of the overall plot
        :param num_cols: The number of columns in the subplot
        :param base64_encode: If True, encode image as base64 in the report HTML. Default is False.
        """
        fig = self.load_imgs_onto_subplot(image_folder_or_paths, figsize=figsize, num_plot_columns=num_cols)
        img_num = len(list(self.report_folder.glob("gallery_image_*")))
        gallery_img_path = self.report_folder / f"gallery_image_{img_num}.png"
        fig.savefig(str(gallery_img_path))
        self.add_images([gallery_img_path], base64_encode=base64_encode)

    def add_plot(
        self, plot_path: Optional[Path] = None, fig: Optional[plt.Figure] = None, fig_title: Optional[str] = None
    ) -> None:
        """
        Add a plot to your report. The plot can either be passed as a [matplotlib Figure object](
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html), or as a path to
        a saved plot file.

        :param plot_path: Optional path to a saved plot file
        :param fig: Optional matplotlib Figure object
        :param fig_title: Optionally provide a title to use as the saved figure filename
        """
        if fig is not None:
            # save the plot
            if fig._suptitle is not None:
                plot_title = fig._suptitle.get_text().replace(" ", "_")
            elif len(fig.texts) > 0:
                plot_title = fig.texts[0].get_text().replace(" ", "_")
            elif fig_title is not None:
                plot_title = fig_title.replace(" ", "_")
            else:
                plot_title = f"plot_{datetime.now().strftime('%Y%m%d%H%M%S')}"
            plot_path = (self.report_folder / plot_title).with_suffix(".png")
            fig.tight_layout()
            fig.savefig(str(plot_path), bbox_inches='tight', dpi=150)
        assert plot_path is not None  # for pyright
        self.add_images([plot_path])

    def read_config_yaml(self, report_config_path: Path) -> OrderedDict:
        """
        Load a report description from a yaml file and use it create an HTML report. The yaml file must as a minimum
        contain the section "report_contents", which takes a list of tuples of component types and their paths.
        For example, to add an image, the tuple might look like ("image", "path/to/image"). For a table, it might
        look like ("table", "path/to/csv/file"). The syntax for text is different, as the second component in
        the tuple is not a path, but the text itself. E.g. ("text", "A section heading"). Provided that one or more
        of these elements is found, self.template and self.render_kwargs will be updated

        :param report_config_path: The path to the .yml file contianing the report description
        :raises ValueError: If a tuple in the report_contents section as a first entry other than "image", "table"
            or "text"
        :return: An OrderedDict representing the contents of the yaml file
        """
        # TODO: add option to overwrite report title with entry here
        assert report_config_path.suffix == ".yml", f"Expected a .yml file but found {report_config_path.suffix}"
        with open(report_config_path, "r") as f_path:
            yaml = YAML(typ='safe', pure=True)
            yaml_contents = yaml.load(f_path)

        return yaml_contents

    def add_yaml_contents_to_report(self, yaml_contents: OrderedDict) -> None:
        """
        Given an OrderdDict of report contents, add these to the report. report_contents is an OrderedDict
        object containing the key "report_contents", whose value is a list of dictionaries. Each dictionary
        has at least 2 entries: firstly, a "type" - e.g. "image", "image_gallery", "table" or "text". The
        second entry each dictionary has is "value". In the case of "text" types, this is the
        text to render. In the case of "image", "image_gallery" or "table" types, this is a path to the content.
        These can either be paths to a single image/csv file, or to a folder containing multiple images/ csv
        files. If multiple, they will be embedded successively, unless the type "image_gallery" has been
        specified

        Some entries may have additional entries, such as "num_columns" or "figsize" for an image_gallery.

        E.g.
        {"table_contents" :
            [
                {"type": "image", "value": "<path/to/image/file_or_folder>"},
                {"type": "table", "value": "<path/to/csv_file_or_folder>"},
                {"type": "text", "value": "A subsection header"}
            ]
        }

        The attribtues "template" and "render_kwargs" will be updated, as the respective methods add_image, add_table
        and add_text are called as necessary.

        :param yaml_contents: An OrderedDict containing at least a "report_contents" section
        :raises ValueError: If the first entry in a row of report_contents is something other than image, table or text
        """
        report_contents = yaml_contents[REPORT_CONTENTS_KEY]
        for component in report_contents:
            component_type = component[ReportComponentKey.TYPE.value]
            component_val = component[ReportComponentKey.VALUE.value]
            base64_encode = component["base64_encode"] if "base64_encode" in component else False
            figsize = component["figsize"] if "figsize" in component else DEFAULT_FIGSIZE
            num_cols = component["num_cols"] if "num_cols" in component else DEFAULT_NUM_COLS
            dir_or_paths = [Path(x) for x in str(component_val).split(",")]
            if component_type == ReportComponentKey.TABLE.value:
                self.add_tables(table_paths_or_dir=dir_or_paths)
            elif component_type == ReportComponentKey.IMAGE.value:
                self.add_images(dir_or_paths, base64_encode=base64_encode)
            elif component_type == ReportComponentKey.IMAGE_GALLERY.value:
                self.add_image_gallery(dir_or_paths, figsize=figsize, num_cols=num_cols, base64_encode=base64_encode)
            elif component_type == ReportComponentKey.TEXT.value:
                self.add_text(component_val)
            else:
                raise ValueError("Key must either equal table, image or text")

    def download_report_contents_from_aml(
        self, run_id: str, report_contents: List[Dict[str, Any]], hyperdrive_hyperparam_name: str = ''
    ) -> List[Dict[str, Any]]:
        """
        Downloads report contents (images, csv files etc, as specified in the ) from Azure ML Runs. If the
        run_id provided represents an AML HyperDrive run, will attempt to download each of the specified paths
        in report_contents from each of the child runs. These will be saved into separate folders in the
        report folder, and the return value will include a comma separated string of the paths to each
        downloaded file. Note that to retrieve these runs, you must provide a value for `hyperdrive_hyperparam_name`
        - i.e. the name of a hyperparameter that was sampled over during this run.

        If not a HyperDrive run, a single file or folder will be downloaded,


        :param run_id: A string representing the run id of the AML run from which to download files
        :param report_contents: A list of dictionaries, where the "type" entry in each is the report content type
            (image, table, text) and the "value" entry is either a path to where the file lives in your DataStore,
            or else it is a string to be added to the report.
        :param hyperdrive_hyperparam_name: If the run is a hyperdrive run, specify the name of one of the
            hyperparameters that was sampled here. This is to ensure files are downloaded into logically-named
            folders.
        :return: An updated list of report contents, with paths replaced by the downloaded file paths where
            applicable.
        """
        # If this is a hyperdrive run, download files for each of its children
        run = get_aml_run_from_run_id(run_id)

        updated_report_contents = []
        for component in report_contents:
            component_type = component[ReportComponentKey.TYPE.value]
            component_val = component[ReportComponentKey.VALUE.value]
            # If the component is text, we don't need to download anything
            if component_type == ReportComponentKey.TEXT.value:
                updated_report_contents.append(
                    {ReportComponentKey.TYPE.value: component_type, ReportComponentKey.VALUE.value: component_val}
                )
            else:
                if run.type == "hyperdrive":
                    artifact_paths = download_files_from_hyperdrive_children(
                        run,
                        component_val,
                        self.report_folder,
                        hyperparam_name=hyperdrive_hyperparam_name,
                    )
                    full_artifact_path = ",".join(artifact_paths)
                else:
                    full_artifact_path = str(self.report_folder / component_val)
                    download_files_from_run_id(run_id, self.report_folder, prefix=component_val)

                updated_component = {
                    ReportComponentKey.TYPE.value: component_type,
                    ReportComponentKey.VALUE.value: full_artifact_path,
                }

                # add back any other entries such as figsize, num_columns etc
                additional_keys = set(component.keys()).difference(set(updated_component.keys()))
                for k in additional_keys:
                    updated_component[k] = component[k]
                updated_report_contents.append(updated_component)

        return updated_report_contents

    def render(self, save_html: bool = True) -> None:
        """
        Render the HTML report template with the keyword arguments that have been collected thorughout report
        generation. Unless save_html is False, creates a file at self.report_path_html and writes the HTML
        content to it

        :param save_html: Whether to save the HTML to file
        """
        new_env = self.env.from_string(self.template)
        subs: str = new_env.render(**self.render_kwargs)
        self.report_html = subs
        # write the substitution to a file
        if save_html:
            with open(self.report_path_html, 'w') as f_path:
                f_path.write(subs)

    def zip_report_folder(self) -> Path:
        """
        Zip the report folder at the location '<report_folder>.zip', preserving the directory structure.
        Returns the path to the zipped folder

        :return: The path to the zipped folder
        """
        import zipfile

        report_files = self.report_folder.rglob("*.*")
        zipped_folder_path = self.report_folder.with_suffix(".zip")
        with zipfile.ZipFile(zipped_folder_path, "w") as zipped_folder:
            for report_file in report_files:
                zipped_folder.write(report_file, arcname=report_file.relative_to(self.report_folder))
        print(f"Zipped folder path: {str(zipped_folder_path)}")
        return zipped_folder_path



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from __future__ import annotations

from typing import Dict, List, Optional

from torch.optim.lr_scheduler import CosineAnnealingLR, ExponentialLR, LambdaLR, MultiStepLR, StepLR, _LRScheduler
from torch.optim.optimizer import Optimizer

from health_ml.deep_learning_config import LRSchedulerType, LRWarmUpType, OptimizerParams


def get_current_learning_rates(optimizer: Optimizer) -> List[float]:
    """
    Reads the current values of the learning rate(s) for all parameter groups from the optimizer.
    """
    return [group['lr'] for group in optimizer.param_groups]


class LinearWarmUp(_LRScheduler):
    """
    Implements linear warmup up to a given initial learning rate.
    """

    def __init__(self, optimizer: Optimizer, warmup_epochs: int, final_lr: float, last_epoch: int = -1):
        if warmup_epochs < 0:
            raise ValueError("The number of warmup epochs must be >= 0.")
        self.warmup_epochs = warmup_epochs
        self.final_lr = final_lr
        self.last_epoch = last_epoch
        super().__init__(optimizer, last_epoch)

    def warmup_multiplier(self) -> float:
        if self.warmup_epochs <= 0:
            return 1.0
        if self.last_epoch >= self.warmup_epochs:
            return 1.0
        return (self.last_epoch + 1) / (self.warmup_epochs + 1)

    def get_lr(self) -> List[float]:  # type: ignore
        return [self.final_lr * self.warmup_multiplier()]


class PolynomialLR:
    def __init__(self, gamma: float, l_rate: float, min_l_rate: float, epochs_after_warmup: int) -> None:
        self.gamma = gamma
        self.l_rate = l_rate
        self.min_l_rate = min_l_rate
        self.epochs_after_warmup = epochs_after_warmup

    def get_lr(self, epoch: int) -> float:
        x = self.min_l_rate / self.l_rate
        return (1 - x) * ((1.0 - float(epoch) / self.epochs_after_warmup) ** self.gamma) + x


class SchedulerWithWarmUp(_LRScheduler):
    """
    LR Scheduler which runs a warmup schedule (linear ramp-up) for a few iterations, and then switches to one
    of the normal schedulers.
    """

    def __init__(self, args: OptimizerParams, optimizer: Optimizer, num_epochs: int, last_epoch: int = -1):
        self.optimizer = optimizer
        self.last_epoch = last_epoch
        self.num_epochs = num_epochs
        self.warmup_epochs = 0 if args.l_rate_warmup == LRWarmUpType.NoWarmUp else args.l_rate_warmup_epochs
        self._scheduler = self.get_scheduler(args)
        # This must be called after self.get_scheduler, because we want the optimizer to have the learning rate
        # guided by the warmup schedule
        self._warmup = LinearWarmUp(
            optimizer, warmup_epochs=self.warmup_epochs, final_lr=args.l_rate, last_epoch=last_epoch
        )
        self._last_lr = get_current_learning_rates(optimizer)
        self.min_l_rate = args.min_l_rate
        super().__init__(optimizer, last_epoch)

    def get_scheduler(self, args: OptimizerParams) -> _LRScheduler:
        """
        Create the LR scheduler that will be used after warmup, based on the config params.
        """
        scheduler: _LRScheduler
        epochs_after_warmup = self.num_epochs - self.warmup_epochs
        if args.l_rate_scheduler == LRSchedulerType.Exponential:
            scheduler = ExponentialLR(
                optimizer=self.optimizer, gamma=args.l_rate_exponential_gamma, last_epoch=self.last_epoch
            )
        elif args.l_rate_scheduler == LRSchedulerType.Step:
            scheduler = StepLR(
                optimizer=self.optimizer,
                step_size=args.l_rate_step_step_size,
                gamma=args.l_rate_step_gamma,
                last_epoch=self.last_epoch,
            )
        elif args.l_rate_scheduler == LRSchedulerType.MultiStep:
            assert args.l_rate_multi_step_milestones is not None
            scheduler = MultiStepLR(
                optimizer=self.optimizer,
                milestones=args.l_rate_multi_step_milestones,
                gamma=args.l_rate_multi_step_gamma,
                last_epoch=self.last_epoch,
            )
        elif args.l_rate_scheduler == LRSchedulerType.Polynomial:
            polynomial_lr = PolynomialLR(
                gamma=args.l_rate_polynomial_gamma,
                l_rate=args.l_rate,
                min_l_rate=args.min_l_rate,
                epochs_after_warmup=epochs_after_warmup,
            )
            scheduler = LambdaLR(optimizer=self.optimizer, lr_lambda=polynomial_lr.get_lr, last_epoch=self.last_epoch)
        elif args.l_rate_scheduler == LRSchedulerType.Cosine:
            scheduler = CosineAnnealingLR(
                optimizer=self.optimizer, T_max=epochs_after_warmup, eta_min=args.min_l_rate, last_epoch=self.last_epoch
            )
        else:
            raise ValueError("Unknown learning rate scheduler {}".format(args.l_rate_scheduler))
        return scheduler

    def state_dict(self) -> Dict:
        """
        Returns a dictionary with all the values in this objects __dict__.
        It creates the dictionary entry for variables "_scheduler" and "_warmup_scheduler" separately, by calling
        state_dict for these variables.
        The state dict does not include the state of the optimizer.
        """
        state_dict = {
            key: val
            for key, val in self.__dict__.items()
            if key != "_scheduler" and key != "optimizer" and key != "_warmup"
        }

        state_dict['_scheduler'] = self._scheduler.state_dict()
        state_dict['_warmup'] = self._warmup.state_dict()
        return state_dict

    def load_state_dict(self, state_dict: Dict) -> None:
        """
        Initializes the current object with values from state_dict.
        Initializes variables "_scheduler" and "_warmup_scheduler" separately, by calling load_state_dict
        for these variables.
        """
        top_level = {key: val for key, val in state_dict.items() if key != "_scheduler" and key != "_warmup"}
        self.__dict__.update(top_level)
        self._scheduler.load_state_dict(state_dict["_scheduler"])
        self._warmup.load_state_dict(state_dict["_warmup"])

    def step(self, epoch: Optional[int] = None) -> None:
        # self.step() is called in the _LRScheduler.__init__, as the very last operation, when self.last_epoch == -1
        # Inside of the default implementation of self.step, it calls
        # self.last_epoch += 1
        # values = self.get_lr()
        # The values are then set in the optimizer, and stored in self._last_lr
        if epoch is not None:
            raise ValueError("Calling scheduler.step with an epoch argument will be deprecated.")
        # self.step is called from within the base class constructor, _LRScheduler.__init__
        # The scheduler itself has already been initialized, and scheduler.step has also been called already in
        # the respective constructor. Avoid calling it again here.
        if self.last_epoch != -1:
            if self.last_epoch < self._warmup.warmup_epochs:
                self._warmup.step()
            else:
                self._scheduler.step()
        self.last_epoch += 1
        self._last_lr = get_current_learning_rates(self.optimizer)

    def get_last_lr(self) -> List[float]:
        return self._last_lr



#! /usr/bin/env python

#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import argparse
import contextlib
from datetime import datetime
import logging
import os
import traceback
import param
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

# Add hi-ml packages to sys.path so that AML can find them if we are using the runner directly from the git repo
himl_root = Path(__file__).resolve().parent.parent.parent.parent
folders_to_add = [himl_root / "hi-ml" / "src", himl_root / "hi-ml-azure" / "src", himl_root / "hi-ml-cpath" / "src"]
for folder in folders_to_add:
    if folder.is_dir():
        sys.path.insert(0, str(folder))

from health_azure import AzureRunInfo, submit_to_azure_if_needed  # noqa: E402
from health_azure.argparsing import create_argparser, parse_arguments, ParserResult, apply_overrides  # noqa: E402
from health_azure.amulet import prepare_amulet_job, is_amulet_job  # noqa: E402
from health_azure.datasets import create_dataset_configs  # noqa: E402
from health_azure.himl import DEFAULT_DOCKER_BASE_IMAGE, OUTPUT_FOLDER  # noqa: E402
from health_azure.logging import logging_to_stdout  # noqa: E402
from health_azure.paths import is_himl_used_from_git_repo  # noqa: E402
from health_azure.utils import (  # noqa: E402
    ENV_LOCAL_RANK,
    ENV_NODE_RANK,
    is_local_rank_zero,
    is_running_in_azure_ml,
    set_environment_variables_for_multi_node,
    filter_v2_input_output_args,
    is_global_rank_zero,
)

from health_ml.eval_runner import EvalRunner  # noqa: E402
from health_ml.experiment_config import DEBUG_DDP_ENV_VAR, ExperimentConfig, RunnerMode  # noqa: E402
from health_ml.lightning_container import LightningContainer  # noqa: E402
from health_ml.runner_base import RunnerBase  # noqa: E402
from health_ml.training_runner import TrainingRunner  # noqa: E402
from health_ml.utils import fixed_paths  # noqa: E402
from health_ml.utils.logging import ConsoleAndFileOutput  # noqa: E402
from health_ml.utils.common_utils import check_conda_environment, choose_conda_env_file, is_linux  # noqa: E402
from health_ml.utils.config_loader import ModelConfigLoader  # noqa: E402
from health_ml.utils import health_ml_package_setup  # noqa: E402


# We change the current working directory before starting the actual training. However, this throws off starting
# the child training threads because sys.argv[0] is a relative path when running in AzureML. Turn that into an absolute
# path.
runner_path = Path(sys.argv[0])
sys.argv[0] = str(runner_path.resolve())


def initialize_rpdb() -> None:
    """
    On Linux only, import and initialize rpdb, to enable remote debugging if necessary.
    """
    # rpdb signal trapping does not work on Windows, as there is no SIGTRAP:
    if not is_linux():
        return
    import rpdb

    rpdb_port = 4444
    rpdb.handle_trap(port=rpdb_port)
    # For some reason, os.getpid() does not return the ID of what appears to be the currently running process.
    logging.info(
        "rpdb is handling traps. To debug: identify the main runner.py process, then as root: "
        f"kill -TRAP <process_id>; nc 127.0.0.1 {rpdb_port}"
    )


def create_runner_parser() -> argparse.ArgumentParser:
    """
    Creates a commandline parser, that understands all necessary arguments for training a model

    :return: An instance of ArgumentParser with args from ExperimentConfig added
    """
    config = ExperimentConfig()
    parser = create_argparser(config)
    return parser


class Runner:
    """
    This class contains the high-level logic to start a training run: choose a model configuration by name,
    submit to AzureML if needed, or otherwise start the actual training and test loop.

    :param project_root: The root folder that contains all of the source code that should be executed.
    """

    def __init__(self, project_root: Path):
        self.project_root = project_root
        self.experiment_config: ExperimentConfig = ExperimentConfig()
        self.lightning_container: LightningContainer = None  # type: ignore
        # This field stores the TrainingRunner object that has been created in the most recent call to the run() method.
        self.ml_runner: Optional[RunnerBase] = None

    def parse_and_load_model(self) -> ParserResult:
        """
        Parses the command line arguments, and creates configuration objects for the model itself, and for the
        Azure-related parameters. Sets self.experiment_config to its proper values. Returns the
        parser output from parsing the model commandline arguments.

        :return: ParserResult object containing args, overrides and settings
        """
        # Filter out any args for passing inputs and outputs to scripts with AML SDK v2
        args = sys.argv[1:]
        filtered_args = filter_v2_input_output_args(args)

        parser1 = create_runner_parser()
        parser1_result = parse_arguments(parser1, args=filtered_args)
        experiment_config = ExperimentConfig(**parser1_result.args)

        from health_azure.logging import logging_stdout_handler  # noqa: E402

        if logging_stdout_handler is not None and experiment_config.log_level is not None:
            print(f"Setting custom logging level to {experiment_config.log_level}")
            logging_stdout_handler.setLevel(experiment_config.log_level.value)
        self.experiment_config = experiment_config
        if not experiment_config.model:
            raise ValueError("Parameter 'model' needs to be set to specify which model to run.")
        model_config_loader: ModelConfigLoader = ModelConfigLoader()
        # Create the model as per the "model" commandline option. This is a LightningContainer.
        container = model_config_loader.create_model_config_from_name(model_name=experiment_config.model)

        # parse overrides and apply
        assert isinstance(container, param.Parameterized)
        parser2 = create_argparser(container)
        # For each parser, feed in the unknown settings from the previous parser. All commandline args should
        # be consumed by name, hence fail if there is something that is still unknown.
        parser2_result = parse_arguments(parser2, args=parser1_result.unknown, fail_on_unknown_args=True)
        # Apply the commandline overrides and validate. Any parameters specified on the commandline should have
        # higher priority than what is done in the model variant
        assert isinstance(container, LightningContainer)
        container.set_model_variant(experiment_config.model_variant)
        apply_overrides(container, parser2_result.overrides)  # type: ignore
        container.validate()

        self.lightning_container = container

        return parser2_result

    def validate(self) -> None:
        """
        Runs sanity checks on the whole experiment.
        """
        if not self.experiment_config.submit_to_azure_ml:
            if self.lightning_container.hyperdrive:
                raise ValueError(
                    "HyperDrive for hyperparameters tuning is only supported when submitting the job to "
                    "AzureML. You need to specify a compute cluster with the argument --cluster."
                )
            if self.lightning_container.is_crossvalidation_parent_run and not is_amulet_job():
                raise ValueError(
                    "Cross-validation is only supported when submitting the job to AzureML."
                    "You need to specify a compute cluster with the argument --cluster."
                )

    def additional_run_tags(self, script_params: List[str]) -> Dict[str, str]:
        """
        Gets the set of tags that will be added to the AzureML run as metadata.

        :param script_params: The commandline arguments used to invoke the present script.
        """
        return {
            "commandline_args": " ".join(script_params),
            "tag": self.lightning_container.tag,
            **self.lightning_container.get_additional_aml_run_tags(),
        }

    def additional_environment_variables(self) -> Dict[str, str]:
        return {
            DEBUG_DDP_ENV_VAR: self.experiment_config.debug_ddp.value,
            **self.lightning_container.get_additional_environment_variables(),
        }

    def run(self) -> Tuple[LightningContainer, AzureRunInfo]:
        """
        The main entry point for training and testing models from the commandline. This chooses a model to train
        via a commandline argument, runs training or testing, and writes all required info to disk and logs.

        :return: a tuple of the LightningContainer object and an AzureRunInfo containing all information about
            the present run (whether running in AzureML or not)
        """
        # Suppress the logging from all processes but the one for GPU 0 on each node, to make log files more readable
        log_level = logging.INFO if is_local_rank_zero() else logging.ERROR
        logging_to_stdout(log_level)
        # When running in Azure, also output logging to a file. This can help in particular when jobs
        # get preempted, but we don't get access to the logs from the previous incarnation of the job
        initialize_rpdb()
        self.parse_and_load_model()
        self.validate()
        azure_run_info = self.submit_to_azureml_if_needed()
        self.run_in_situ(azure_run_info)
        return self.lightning_container, azure_run_info

    def submit_to_azureml_if_needed(self) -> AzureRunInfo:
        """
        Submit a job to AzureML, returning the resulting Run object, or exiting if we were asked to wait for
        completion and the Run did not succeed.

        :return: an AzureRunInfo object containing all of the details of the present run. If AzureML is not
            specified, the attribute 'run' will None, but the object still contains helpful information
            about datasets etc
        """
        root_folder = self.project_root
        entry_script = Path(sys.argv[0]).resolve()
        script_params = sys.argv[1:]

        environment_variables = self.additional_environment_variables()
        local_datasets = self.lightning_container.local_datasets
        all_local_datasets = [Path(p) for p in local_datasets] if len(local_datasets) > 0 else []
        # When running in AzureML, respect the commandline flag for mounting. Outside of AML, we always mount
        # datasets to be quicker.
        use_mounting = self.experiment_config.mount_in_azureml if self.experiment_config.submit_to_azure_ml else True
        input_datasets = create_dataset_configs(
            all_azure_dataset_ids=self.lightning_container.azure_datasets,
            all_dataset_mountpoints=self.lightning_container.dataset_mountpoints,
            all_local_datasets=all_local_datasets,  # type: ignore
            datastore=self.lightning_container.datastore,
            use_mounting=use_mounting,
        )

        if self.experiment_config.submit_to_azure_ml and not is_running_in_azure_ml():
            if self.experiment_config.strictly_aml_v1:
                hyperdrive_config = self.lightning_container.get_hyperdrive_config()
                hyperparam_args = None
            else:
                hyperparam_args = self.lightning_container.get_hyperparam_args()
                hyperdrive_config = None

            env_file = choose_conda_env_file(env_file=self.experiment_config.conda_env)
            logging.info(f"Using this Conda environment definition: {env_file}")
            check_conda_environment(env_file)

            azure_run_info = submit_to_azure_if_needed(
                entry_script=entry_script,
                snapshot_root_directory=root_folder,
                script_params=script_params,
                conda_environment_file=env_file,
                compute_cluster_name=self.experiment_config.cluster,
                environment_variables=environment_variables,
                experiment_name=self.lightning_container.effective_experiment_name,
                input_datasets=input_datasets,  # type: ignore
                num_nodes=self.experiment_config.num_nodes,
                wait_for_completion=self.experiment_config.wait_for_completion,
                max_run_duration=self.experiment_config.max_run_duration,
                ignored_folders=[],
                submit_to_azureml=self.experiment_config.submit_to_azure_ml,
                docker_base_image=DEFAULT_DOCKER_BASE_IMAGE,
                docker_shm_size=self.experiment_config.docker_shm_size,
                hyperdrive_config=hyperdrive_config,
                hyperparam_args=hyperparam_args,
                display_name=self.lightning_container.tag,
                tags=self.additional_run_tags(script_params),
                strictly_aml_v1=self.experiment_config.strictly_aml_v1,
                identity_based_auth=self.experiment_config.identity_based_auth,
            )
        else:
            azure_run_info = submit_to_azure_if_needed(
                input_datasets=input_datasets,  # type: ignore
                submit_to_azureml=False,
                environment_variables=environment_variables,
                strictly_aml_v1=self.experiment_config.strictly_aml_v1,
            )
        if azure_run_info.run:
            # This code is only reached inside Azure. Set display name again - this will now affect
            # Hypdrive child runs (for other jobs, this has already been done after submission)
            suffix = None
            if self.lightning_container.is_crossvalidation_child_run:
                suffix = f"crossval {self.lightning_container.crossval_index}"
            elif self.lightning_container.different_seeds > 0:
                suffix = f"seed {self.lightning_container.random_seed}"
            if suffix:
                current_name = self.lightning_container.tag or azure_run_info.run.display_name
                azure_run_info.run.display_name = f"{current_name} {suffix}"
        # submit_to_azure_if_needed calls sys.exit after submitting to AzureML. We only reach this when running
        # the script locally or in AzureML.
        return azure_run_info

    def run_in_situ(self, azure_run_info: AzureRunInfo) -> None:
        """
        Actually run the AzureML job; this method will typically run on an Azure VM.

        :param azure_run_info: Contains all information about the present run in AzureML, in particular where the
        datasets are mounted.
        """
        health_ml_package_setup()
        prepare_amulet_job()

        # Add tags and arguments to Amulet runs
        if is_amulet_job():
            assert azure_run_info.run is not None
            azure_run_info.run.set_tags(self.additional_run_tags(sys.argv[1:]))

        if self.experiment_config.mode == RunnerMode.TRAIN:
            # Set environment variables for multi-node training if needed. This function will terminate early
            # if it detects that it is not in a multi-node environment.
            if self.experiment_config.num_nodes > 1:
                set_environment_variables_for_multi_node()
            self.ml_runner = TrainingRunner(
                experiment_config=self.experiment_config,
                container=self.lightning_container,
                project_root=self.project_root,
            )
        elif self.experiment_config.mode == RunnerMode.EVAL_FULL:
            self.ml_runner = EvalRunner(
                experiment_config=self.experiment_config,
                container=self.lightning_container,
                project_root=self.project_root,
            )
        else:
            raise ValueError(f"Unknown mode {self.experiment_config.mode}")
        self.ml_runner.validate()
        self.ml_runner.run_and_cleanup(azure_run_info)


def run(project_root: Path) -> Tuple[LightningContainer, AzureRunInfo]:
    """
    The main entry point for training and testing models from the commandline. This chooses a model to train
    via a commandline argument, runs training or testing, and writes all required info to disk and logs.

    :param project_root: The root folder that contains all of the source code that should be executed.
    :return: If submitting to AzureML, returns the model configuration that was used for training,
        including commandline overrides applied (if any). For details on the arguments, see the constructor of Runner.
    """
    if is_global_rank_zero():
        print(f"Project root: {project_root}")
    return Runner(project_root).run()


def create_logging_filename() -> Path:
    """Creates a file name for console logs, based on the current time and the DDP rank.
    The filename is timestamped to seconds level, so that we also get a full history of all
    low-priority preemptions, because each restart will create a logfile afresh.

    :return: A full path to a file for console logs.
    """
    rank = os.getenv(ENV_LOCAL_RANK, "0")
    node = os.getenv(ENV_NODE_RANK, "0")
    timestamp = datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")
    cwd = Path.cwd()
    logging_filename = cwd
    # DDP subprocesses may already be running in the "outputs" folder. Add the "outputs" hence only if necessary.
    if cwd.name != OUTPUT_FOLDER:
        logging_filename = logging_filename / Path(OUTPUT_FOLDER)
    logging_filename = logging_filename / "console_logs" / f"logging_{timestamp}_node{node}_rank{rank}.txt"
    logging_filename.parent.mkdir(parents=True, exist_ok=True)
    print(f"Rank {rank}: Redirecting all console logs to {logging_filename}")
    return logging_filename


def run_with_logging(project_root: Path) -> Tuple[LightningContainer, AzureRunInfo]:
    """
    Start the main main entry point for training and testing models from the commandline.
    When running in Azure, this method also redirects the stdout stream, so that all console output is visible both on
    the console and stored in a file. The filename is timestamped and contains the DDP rank of the current process.

    :param project_root: The root folder that contains all of the source code that should be executed.
    :return: If submitting to AzureML, returns the model configuration that was used for training,
        including commandline overrides applied (if any). For details on the arguments, see the constructor of Runner.
    """
    if is_running_in_azure_ml():
        logging_filename = create_logging_filename()
        with logging_filename.open("w") as logging_file:
            console_and_file = ConsoleAndFileOutput(logging_file)
            with contextlib.redirect_stdout(console_and_file):
                try:
                    return run(project_root)
                except:  # noqa
                    # Exceptions would only be printed to the console at the very top level, and would not be visible
                    # in the log file. Hence, write here specifically.
                    traceback.print_exc(file=logging_file)
                    raise
                finally:
                    logging_file.flush()
    return run(project_root)


def main() -> None:
    project_root = fixed_paths.repository_root_directory() if is_himl_used_from_git_repo() else Path.cwd()
    run_with_logging(project_root)


if __name__ == '__main__':
    main()



from pytorch_lightning import LightningDataModule

from health_azure.logging import logging_section
from health_ml.runner_base import RunnerBase


class EvalRunner(RunnerBase):
    """A class to run the evaluation of a model on a new dataset. The initialization logic is taken from the base
    class `RunnerBase`.
    """

    def validate(self) -> None:
        """Checks if the fields of the class are set up correctly."""
        if self.container.src_checkpoint is None or self.container.src_checkpoint.checkpoint == "":
            raise ValueError(
                "To use model evaluation, you need to provide a checkpoint to use, via the --src_checkpoint argument."
            )

    def run(self) -> None:
        """Start the core workflow that the class implements: Initialize a PL Trainer object and use that to run
        inference on the inference dataset."""
        self.container.outputs_folder.mkdir(exist_ok=True, parents=True)
        self.init_inference()
        with logging_section("Model inference"):
            self.run_inference()

    def get_data_module(self) -> LightningDataModule:
        """Reads the evaluation data module from the underlying container."""
        return self.container.get_eval_data_module()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

import numpy as np
from scipy import ndimage


@dataclass(frozen=True)
class Box:
    """Utility class representing rectangular regions in 2D images.

    :param x: Horizontal coordinate of the top-left corner.
    :param y: Vertical coordinate of the top-left corner.
    :param w: Box width.
    :param h: Box height.
    :raises ValueError: If either `w` or `h` are <= 0.
    """

    x: int
    y: int
    w: int
    h: int

    def __post_init__(self) -> None:
        if self.w <= 0:
            raise ValueError(f"Width must be strictly positive, received {self.w}")
        if self.h <= 0:
            raise ValueError(f"Height must be strictly positive, received {self.w}")

    def __add__(self, shift: Sequence[int]) -> 'Box':
        """Translates the box's location by a given shift.

        :param shift: A length-2 sequence containing horizontal and vertical shifts.
        :return: A new box with updated `x = x + shift[0]` and `y = y + shift[1]`.
        :raises ValueError: If `shift` does not have two elements.
        """
        if len(shift) != 2:
            raise ValueError("Shift must be two-dimensional")
        return Box(x=self.x + shift[0], y=self.y + shift[1], w=self.w, h=self.h)

    def __mul__(self, factor: float) -> 'Box':
        """Scales the box by a given factor, e.g. when changing resolution.

        :param factor: The factor by which to multiply the box's location and dimensions.
        :return: The updated box, with location and dimensions rounded to `int`.
        """
        return Box(x=int(self.x * factor), y=int(self.y * factor), w=int(self.w * factor), h=int(self.h * factor))

    def __rmul__(self, factor: float) -> 'Box':
        """Scales the box by a given factor, e.g. when changing resolution.

        :param factor: The factor by which to multiply the box's location and dimensions.
        :return: The updated box, with location and dimensions rounded to `int`.
        """
        return self * factor

    def __truediv__(self, factor: float) -> 'Box':
        """Scales the box by a given factor, e.g. when changing resolution.

        :param factor: The factor by which to divide the box's location and dimensions.
        :return: The updated box, with location and dimensions rounded to `int`.
        """
        return self * (1.0 / factor)

    def add_margin(self, margin: int) -> 'Box':
        """Adds a symmetric margin on all sides of the box.

        :param margin: The amount by which to enlarge the box.
        :return: A new box enlarged by `margin` on all sides.
        """
        return Box(x=self.x - margin, y=self.y - margin, w=self.w + 2 * margin, h=self.h + 2 * margin)

    def clip(self, other: 'Box') -> Optional['Box']:
        """Clips a box to the interior of another.

        This is useful to constrain a region to the interior of an image.

        :param other: Box representing the new constraints.
        :return: A new constrained box, or `None` if the boxes do not overlap.
        """
        x0 = max(self.x, other.x)
        y0 = max(self.y, other.y)
        x1 = min(self.x + self.w, other.x + other.w)
        y1 = min(self.y + self.h, other.y + other.h)
        try:
            return Box(x=x0, y=y0, w=x1 - x0, h=y1 - y0)
        except ValueError:  # Empty result, boxes don't overlap
            return None

    def to_slices(self) -> Tuple[slice, slice]:
        """Converts the box to slices for indexing arrays.

        For example: `my_2d_array[my_box.to_slices()]`.

        :return: A 2-tuple with vertical and horizontal slices.
        """
        return (slice(self.y, self.y + self.h), slice(self.x, self.x + self.w))

    @staticmethod
    def from_slices(slices: Sequence[slice]) -> 'Box':
        """Converts a pair of vertical and horizontal slices into a box.

        :param slices: A length-2 sequence containing vertical and horizontal `slice` objects.
        :return: A box with corresponding location and dimensions.
        """
        vert_slice, horz_slice = slices
        return Box(
            x=horz_slice.start,
            y=vert_slice.start,
            w=horz_slice.stop - horz_slice.start,
            h=vert_slice.stop - vert_slice.start,
        )


def get_bounding_box(mask: np.ndarray) -> Box:
    """Extracts a bounding box from a binary 2D array.

    :param mask: A 2D array with 0 (or `False`) as background and >0 (or `True`) as foreground.
    :return: The smallest box covering all non-zero elements of `mask`.
    :raises TypeError: When the input mask has more than two dimensions.
    :raises RuntimeError: When all elements in the mask are zero.
    """
    if mask.ndim != 2:
        raise TypeError(f"Expected a 2D array but got an array with shape {mask.shape}")

    slices = ndimage.find_objects(mask > 0)
    if not slices:
        raise RuntimeError("The input mask is empty")
    assert len(slices) == 1

    return Box.from_slices(slices[0])



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
import time
from typing import Any, Optional, Set

import torch
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.utilities.distributed import rank_zero_only
from pytorch_lightning.utilities.types import STEP_OUTPUT


class EpochTimers:
    """
    Contains all information necessary to compute the IO metrics: Epoch times, batch times, loading times.
    """

    def __init__(
        self, max_batch_load_time_seconds: float = 0.5, max_load_time_warnings: int = 3, max_load_time_epochs: int = 5
    ) -> None:
        """
        :param max_batch_load_time_seconds: The maximum expected loading time for a minibatch (given in seconds).
            If the loading time exceeds this threshold, a warning is printed.
        :param max_load_time_warnings: The maximum number of warnings that will be printed per epoch.
        :param max_load_time_epochs: The maximum number of epochs where warnings about the loading time are printed.
        """
        self.max_batch_load_time_seconds = max_batch_load_time_seconds
        self.max_load_time_warnings = max_load_time_warnings
        self.max_load_time_epochs = max_load_time_epochs
        self.load_time_warning_epochs: Set[int] = set()
        self.epoch_start_time: float = 0.0
        self.epoch_end_time: float = 0.0
        self.batch_start_time: float = 0.0
        self.num_load_time_warnings: int = 0
        self.num_load_time_exceeded: int = 0
        self.total_extra_load_time: float = 0.0
        self.total_load_time: float = 0.0
        self.num_batches: int = 0

    def epoch_start(self) -> None:
        """
        Resets all timers to the current time, and all counters to 0. The set of epochs for which warnings about
        load time were produced will not be reset.
        """
        current_time = time.time()
        self.epoch_start_time = current_time
        self.epoch_end_time = current_time
        self.batch_start_time = current_time
        self.num_load_time_warnings = 0
        self.num_load_time_exceeded = 0
        self.total_extra_load_time = 0.0
        self.total_load_time = 0.0
        self.num_batches = 0

    def epoch_end(self) -> None:
        """
        Stores the present time in the epoch_end_time field of the object.
        """
        self.epoch_end_time = time.time()

    @property
    def total_epoch_time(self) -> float:
        """
        Gets the time in seconds between epoch start and epoch end.
        """
        return self.epoch_end_time - self.epoch_start_time

    @property
    def should_warn_in_this_epoch(self) -> bool:
        """
        Returns True if warnings about loading time should be printed in the present epoch. Returns False if
        this warning has been printed already in more than self.max_load_time_epochs epochs.
        """
        return len(self.load_time_warning_epochs) <= self.max_load_time_epochs

    def batch_start(self, batch_index: int, epoch: int, message_prefix: str) -> float:
        """
        Called when a minibatch of data has been loaded. This computes the time it took to load the minibatch
        (computed between now and the end of the previous minibatch)
        and adds it to the internal bookkeeping. If the minibatch loading time exceeds a threshold, then warnings
        are printed (unless too many warnings have been printed already)

        :param message_prefix: A prefix string that is added to all diagnostic output.
        :param epoch: The index of the current epoch.
        :param batch_index: The index of the current minibatch.
        :return: The time it took to load the minibatch, in seconds.
        """
        item_finish_time = time.time()
        item_load_time = item_finish_time - self.batch_start_time
        self.total_load_time += item_load_time
        # Having slow minibatch loading is OK in the very first batch of the every epoch, where processes
        # are spawned. Later, the load time should be zero.
        if batch_index == 0:
            logging.info(f"{message_prefix}: Loaded the first minibatch of data in {item_load_time:0.2f} sec.")
        elif item_load_time > self.max_batch_load_time_seconds:
            self.load_time_warning_epochs.add(epoch)
            self.num_load_time_exceeded += 1
            self.total_extra_load_time += item_load_time
            if self.num_load_time_warnings < self.max_load_time_warnings and self.should_warn_in_this_epoch:
                logging.warning(
                    f"{message_prefix}: Loading minibatch {batch_index} took {item_load_time:0.2f} sec. "
                    "This can mean that there are not enough data loader worker processes, or that there "
                    "is a performance problem in loading. This warning will be printed at most "
                    f"{self.max_load_time_warnings} times in at most {self.max_load_time_epochs} epochs."
                )
                self.num_load_time_warnings += 1
        return item_load_time

    def batch_end(self) -> float:
        """
        Called after a minibatch has been processed (training or validation step completed). Returns the time it took
        to process the current batch (including loading).

        :return: The time it took to process the current batch, in seconds.
        """
        current_time = time.time()
        elapsed = current_time - self.batch_start_time
        self.batch_start_time = current_time
        self.num_batches += 1
        return elapsed


class BatchTimeCallback(Callback):
    """
    This callback provides tools to measure batch loading time and other diagnostic information.
    It prints alerts to the console or to `logging` if the batch loading time is over a threshold for several epochs.
    Metrics for loading time, as well as epoch time, and maximum and average batch processing time are logged to
    the loggers that are set up on the module.
    In distributed training, all logging to the console and to the Lightning loggers will only happen on global rank 0.

    The loading time for a minibatch is estimated by the difference between the start time of a minibatch and the
    end time of the previous minibatch. It will consequently also include other operations that happen between the
    end of a batch and the start of the next one. For example, computationally expensive callbacks could also
    drive up this time.

    Usage example:
        >>> from health_ml.utils import BatchTimeCallback
        >>> from pytorch_lightning import Trainer
        >>> batchtime = BatchTimeCallback(max_batch_load_time_seconds=0.5)
        >>> trainer = Trainer(callbacks=[batchtime])
    """

    EPOCH_TIME = "epoch_time [sec]"
    """The name that is used to log the execution time per epoch """
    BATCH_TIME = "batch_time [sec]"
    """The name that is used to log the execution time per batch."""
    EXCESS_LOADING_TIME = "batch_loading_over_threshold [sec]"
    """The name that is used to log the time spent loading all the batches that exceeding the loading time threshold."""
    METRICS_PREFIX = "timing/"
    """The prefix for all metrics collected by this callback."""
    TRAIN_PREFIX = "train/"
    """The prefix for all metrics collected during training."""
    VAL_PREFIX = "val/"
    """The prefix for all metrics collected during validation."""

    def __init__(
        self, max_batch_load_time_seconds: float = 0.5, max_load_time_warnings: int = 3, max_load_time_epochs: int = 5
    ) -> None:
        """
        :param max_batch_load_time_seconds: The maximum expected loading time for a minibatch (given in seconds).
            If the loading time exceeds this threshold, a warning is printed. The maximum number of such warnings is
            controlled by the other arguments.
        :param max_load_time_warnings: The maximum number of warnings about increased loading time that will be printed
            per epoch. For example, if max_load_time_warnings=3, at most 3 of these warnings will be printed within an
            epoch. The 4th minibatch with loading time over the threshold would not generate any warning anymore.
            If set to 0, no warnings are printed at all.
        :param max_load_time_epochs: The maximum number of epochs where warnings about the loading time are printed.
            For example, if max_load_time_epochs=2, and at least 1 batch with increased loading time is observed in
            epochs 0 and 3, no further warnings about increased loading time would be printed from epoch 4 onwards.
        """
        # Timers for monitoring data loading time
        self.train_timers = EpochTimers(
            max_batch_load_time_seconds=max_batch_load_time_seconds,
            max_load_time_warnings=max_load_time_warnings,
            max_load_time_epochs=max_load_time_epochs,
        )
        self.val_timers = EpochTimers(
            max_batch_load_time_seconds=max_batch_load_time_seconds,
            max_load_time_warnings=max_load_time_warnings,
            max_load_time_epochs=max_load_time_epochs,
        )
        self.module: Optional[LightningModule] = None

    def on_fit_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """
        This is called at the start of training. It stores the model that is being trained, because it will be used
        later to log values.
        """
        self.module = pl_module

    @rank_zero_only
    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.train_timers.epoch_start()

    @rank_zero_only
    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.val_timers.epoch_start()
        # In Lightning, the validation epoch is running "inside" the training. If we get here, it means that training
        # is done for this epoch, even though the on_training_epoch hook has not yet been called.
        self.train_timers.epoch_end()

    @rank_zero_only
    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """
        This is a hook called at the end of a training or validation epoch. In here, we can still write
        metrics to a logger.
        """
        # In validation epochs, mark that it has been completed. Training epochs are marked completed already
        # at the start of the validation epoch.
        self.val_timers.epoch_end()
        # Write all IO stats here, so that the order on the console is Train start, train end, val start, val end.
        self.write_and_log_epoch_time(is_training=True)
        self.write_and_log_epoch_time(is_training=False)

    def on_train_batch_start(
        self,  # type: ignore
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self.batch_start(batch_idx=batch_idx, is_training=True)

    def on_validation_batch_start(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self.batch_start(batch_idx=batch_idx, is_training=False)

    def on_train_batch_end(
        self,  # type: ignore
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self.batch_end(is_training=True)

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self.batch_end(is_training=False)

    @rank_zero_only
    def batch_start(self, batch_idx: int, is_training: bool) -> None:
        """
        Shared code to keep track of minibatch loading times. This is only done on global rank zero.

        :param batch_idx: The index of the current minibatch.
        :param is_training: If true, this has been called from `on_train_batch_start`, otherwise it has been called from
            `on_validation_batch_start`.
        """
        timers = self.get_timers(is_training=is_training)
        assert self.module is not None
        epoch = self.module.current_epoch
        message_prefix = f"Epoch {epoch} {'training' if is_training else 'validation'}"
        timers.batch_start(batch_index=batch_idx, epoch=epoch, message_prefix=message_prefix)

    @rank_zero_only
    def batch_end(self, is_training: bool) -> None:
        """
        Shared code to keep track of minibatch loading times. This is only done on global rank zero.

        :param is_training: If true, this has been called from `on_train_batch_end`, otherwise it has been called from
            `on_validation_batch_end`.
        """
        timers = self.get_timers(is_training=is_training)
        batch_time = timers.batch_end()
        self.log_metric(self.BATCH_TIME + " avg", value=batch_time, is_training=is_training)
        self.log_metric(self.BATCH_TIME + " max", value=batch_time, is_training=is_training, reduce_max=True)

    @rank_zero_only
    def write_and_log_epoch_time(self, is_training: bool) -> None:
        """
        Reads the IO timers for either the training or validation epoch, writes them to the console, and logs the
        time per epoch.

        :param is_training: If True, show and log the data for the training epoch. If False, use the data for the
            validation epoch.
        """
        timers = self.get_timers(is_training=is_training)
        epoch_time_seconds = timers.total_epoch_time
        status = "training" if is_training else "validation"
        assert self.module is not None
        logging.info(
            f"Epoch {self.module.current_epoch} {status} took {epoch_time_seconds:0.2f}sec, of which waiting "
            f"for data took {timers.total_load_time:0.2f} sec total."
        )
        if timers.num_load_time_exceeded > 0 and timers.should_warn_in_this_epoch:
            logging.warning(
                "The dataloaders were not fast enough to always supply the next batch in less than "
                f"{timers.max_batch_load_time_seconds:0.2f}sec."
            )
            logging.warning(
                f"In this epoch, {timers.num_load_time_exceeded} out of {timers.num_batches} batches exceeded the load "
                f"time threshold. Total loading time for the slow batches was {timers.total_extra_load_time:0.2f}sec."
            )
        # This metric is only written at rank zero, and hence must not be synchronized across workers. If attempted,
        # training will get stuck.
        self.log_metric(self.EPOCH_TIME, value=epoch_time_seconds, is_training=is_training)
        self.log_metric(self.EXCESS_LOADING_TIME, value=timers.total_extra_load_time, is_training=is_training)

    @rank_zero_only
    def log_metric(self, name_suffix: str, value: float, is_training: bool, reduce_max: bool = False) -> None:
        """
        Write a metric given as a name/value pair to the currently trained module. The full name of the metric is
        composed of a fixed prefix "timing/", followed by either "train/" or "val/", and then the given suffix.

        :param name_suffix: The suffix for the logged metric name.
        :param value: The value to log.
        :param is_training: If True, use "train/" in the metric name, otherwise "val/"
        :param reduce_max: If True, use torch.max as the aggregation function for the logged values. If False, use
            torch.mean
        """
        # Metrics are only written at global rank 0, and hence must not be synchronized. Trying to synchronize will
        # block training.
        prefix = self.TRAIN_PREFIX if is_training else self.VAL_PREFIX
        assert self.module is not None
        self.module.log(
            name=self.METRICS_PREFIX + prefix + name_suffix,
            value=value,  # type: ignore
            on_step=False,
            on_epoch=True,
            sync_dist=False,
            reduce_fx=max if reduce_max else torch.mean,
        )

    def get_timers(self, is_training: bool) -> EpochTimers:
        """
        Gets the object that holds all metrics and timers, for either the validation or the training epoch.
        """
        return self.train_timers if is_training else self.val_timers


class TrainingDiagnoticsCallback(Callback):
    """Callback that logs information about the training process."""

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_train_epoch_start on global rank {pl_module.global_rank}")

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_train_epoch_end on global rank {pl_module.global_rank}")

    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_validation_epoch_start on global rank {pl_module.global_rank}")

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_validation_epoch_end on global rank {pl_module.global_rank}")

    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_test_epoch_start on global rank {pl_module.global_rank}")

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_test_epoch_end on global rank {pl_module.global_rank}")

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached on_train_start on global rank {pl_module.global_rank}")

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached from on_train_end on global rank {pl_module.global_rank}")

    def on_validation_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached from on_validation_start on global rank {pl_module.global_rank}")

    def on_validation_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached from on_validation_end on global rank {pl_module.global_rank}")

    def on_test_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached from on_test_start on global rank {pl_module.global_rank}")

    def on_test_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        logging.info(f"Reached from on_test_end on global rank {pl_module.global_rank}")

    def on_train_batch_start(
        self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int, unused: int = 0
    ) -> None:
        logging.info(
            f"Reached from on_train_batch_start on global rank {pl_module.global_rank} for batch_idx {batch_idx}"
        )

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        unused: int = 0,
    ) -> None:
        logging.info(
            f"Reached from on_train_batch_end on global rank {pl_module.global_rank} for batch_idx {batch_idx}"
        )

    def on_validation_batch_start(
        self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int, unused: int = 0
    ) -> None:
        logging.info(
            f"Reached from on_validation_batch_start on global rank {pl_module.global_rank} for batch_idx {batch_idx}"
        )

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        logging.info(
            f"Reached from on_validation_batch_end on global rank {pl_module.global_rank} for batch_idx {batch_idx}"
        )

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        logging.info(f"Reached from on_test_batch_end on global rank {pl_module.global_rank} for batch_idx {batch_idx}")

    def on_test_batch_start(
        self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int, unused: int = 0
    ) -> None:
        logging.info(
            f"Reached from on_test_batch_start on global rank {pl_module.global_rank} for batch_idx {batch_idx}"
        )
